from pytest import approx
from pybliometrics.scival import TopicLookupMetrics, init

init()

topic = TopicLookupMetrics("9549", by_year=False, refresh=30)
topics_by_year = TopicLookupMetrics("110, 1438, 1012", by_year=True, refresh=30)
topics_not_by_year = TopicLookupMetrics(["3909", "4374"], by_year=False, refresh=30)
specific_metrics = TopicLookupMetrics("1516, 2848",
                                metric_types=["AuthorCount", "CitationCount"],
                                by_year=False, refresh=30)
non_existing = TopicLookupMetrics("9999999999", by_year=False, refresh=30)

def test_topics():
    """Tests topics property"""
    expected_fields = ("id", "name", "uri", "prominencePercentile", "scholarlyOutput")

    assert len(topic.topics) == 1
    assert topic.topics[0]._fields == expected_fields
    assert topic.topics[0].id == 9549
    assert topic.topics[0].name == "Quantum States and Entanglement in Many-Body Systems"
    assert topic.topics[0].uri == "Topic/9549"
    assert topic.topics[0].prominencePercentile == approx(94, abs=10)
    assert topic.topics[0].scholarlyOutput == approx(822, rel=0.1)

    assert len(topics_by_year.topics) == 3
    assert topics_by_year.topics[0]._fields == expected_fields

    assert len(topics_not_by_year.topics) == 2
    assert topics_not_by_year.topics[0]._fields == expected_fields

    assert len(specific_metrics.topics) == 2
    assert specific_metrics.topics[0]._fields == expected_fields

    assert len(non_existing.topics) == 0

def test_author_count():
    """Test AuthorCount property"""
    expected_fields = ("entity_id", "entity_name", "metric", "year", "value", "percentage", "threshold")

    assert len(topic.AuthorCount) == 1
    assert topic.AuthorCount[0]._fields == expected_fields

    assert len(topics_by_year.AuthorCount) == 15
    assert topics_by_year.AuthorCount[0]._fields == expected_fields
    assert topics_by_year.AuthorCount[0].entity_id == 110
    assert topics_by_year.AuthorCount[0].entity_name == "Photoacoustic Imaging Techniques for Enhanced Tissue Analysis"
    assert topics_by_year.AuthorCount[0].metric == "AuthorCount"
    assert topics_by_year.AuthorCount[0].year == "2020"
    assert topics_by_year.AuthorCount[0].value == approx(1997, rel=0.1)
    assert topics_by_year.AuthorCount[0].percentage is None
    assert topics_by_year.AuthorCount[0].threshold is None

    assert len(topics_not_by_year.AuthorCount) == 2
    assert topics_not_by_year.AuthorCount[0]._fields == expected_fields

    assert len(specific_metrics.AuthorCount) == 2
    assert specific_metrics.AuthorCount[0]._fields == expected_fields

    assert non_existing.AuthorCount is None


def test_citation_count():
    """Test CitationCount property"""
    expected_fields = ("entity_id", "entity_name", "metric", "year", "value", "percentage", "threshold")

    assert len(topic.CitationCount) == 1
    assert topic.CitationCount[0]._fields == expected_fields

    assert len(topics_by_year.CitationCount) == 15
    assert topics_by_year.CitationCount[0]._fields == expected_fields

    assert len(topics_not_by_year.CitationCount) == 2
    assert topics_not_by_year.CitationCount[0]._fields == expected_fields
    assert topics_not_by_year.CitationCount[0].entity_id == 3909
    assert topics_not_by_year.CitationCount[0].entity_name == "Point Cloud Semantic Segmentation"
    assert topics_not_by_year.CitationCount[0].metric == "CitationCount"
    assert topics_not_by_year.CitationCount[0].year == "all"
    assert topics_not_by_year.CitationCount[0].value == approx(84184, rel=0.1)
    assert topics_not_by_year.CitationCount[0].percentage is None
    assert topics_not_by_year.CitationCount[0].threshold is None

    assert len(specific_metrics.CitationCount) == 2
    assert specific_metrics.CitationCount[0]._fields == expected_fields

    assert non_existing.CitationCount is None


def test_core_papers():
    """Test CorePapers property"""
    expected_fields = ('entity_id', 'entity_name', 'publication_id')

    assert len(topic.CorePapers) == 10
    assert topic.CorePapers[0]._fields == expected_fields

    assert topics_by_year.CorePapers is None

    assert len(topics_not_by_year.CorePapers) == 20
    assert topics_not_by_year.CorePapers[0]._fields == expected_fields
    assert topics_not_by_year.CorePapers[0].entity_id == 3909
    assert topics_not_by_year.CorePapers[0].entity_name == "Point Cloud Semantic Segmentation"
    assert topics_not_by_year.CorePapers[0].publication_id == 85092710521

    assert specific_metrics.CorePapers is None

    assert non_existing.CorePapers is None

def test_field_weighted_citation_impact():
    """Test FieldWeightedCitationImpact property"""
    expected_fields = ("entity_id", "entity_name", "metric", "year", "value", "percentage", "threshold")

    assert len(topic.FieldWeightedCitationImpact) == 1
    assert topic.FieldWeightedCitationImpact[0]._fields == expected_fields

    assert len(topics_by_year.FieldWeightedCitationImpact) == 15
    assert topics_by_year.FieldWeightedCitationImpact[0]._fields == expected_fields

    assert len(topics_not_by_year.FieldWeightedCitationImpact) == 2
    assert topics_not_by_year.FieldWeightedCitationImpact[0]._fields == expected_fields
    assert topics_not_by_year.FieldWeightedCitationImpact[0].entity_id == 3909
    assert topics_not_by_year.FieldWeightedCitationImpact[0].entity_name == "Point Cloud Semantic Segmentation"
    assert topics_not_by_year.FieldWeightedCitationImpact[0].metric == "FieldWeightedCitationImpact"
    assert topics_not_by_year.FieldWeightedCitationImpact[0].year == "all"
    assert topics_not_by_year.FieldWeightedCitationImpact[0].value == approx(1.7, rel=0.1)
    assert topics_not_by_year.FieldWeightedCitationImpact[0].percentage is None
    assert topics_not_by_year.FieldWeightedCitationImpact[0].threshold is None

    assert specific_metrics.FieldWeightedCitationImpact is None

    assert non_existing.FieldWeightedCitationImpact is None

def test_institution_count():
    """Test InstitutionCount property"""
    expected_fields = ("entity_id", "entity_name", "metric", "year", "value", "percentage", "threshold")

    assert len(topic.InstitutionCount) == 1
    assert topic.InstitutionCount[0]._fields == expected_fields

    assert len(topics_by_year.InstitutionCount) == 15
    assert topics_by_year.InstitutionCount[0]._fields == expected_fields

    assert len(topics_not_by_year.InstitutionCount) == 2
    assert topics_not_by_year.InstitutionCount[0]._fields == expected_fields
    assert topics_not_by_year.InstitutionCount[0].entity_id == 3909
    assert topics_not_by_year.InstitutionCount[0].entity_name == "Point Cloud Semantic Segmentation"
    assert topics_not_by_year.InstitutionCount[0].metric == "InstitutionCount"
    assert topics_not_by_year.InstitutionCount[0].year == "all"
    assert topics_not_by_year.InstitutionCount[0].value == approx(1551, rel=0.1)
    assert topics_not_by_year.InstitutionCount[0].percentage is None
    assert topics_not_by_year.InstitutionCount[0].threshold is None

    assert specific_metrics.InstitutionCount is None

    assert non_existing.InstitutionCount is None


def test_most_recently_published_papers():
    """Test MostRecentlyPublishedPapers property"""
    expected_fields = ('entity_id', 'entity_name', 'publication_id')

    assert len(topic.MostRecentlyPublishedPapers) == approx(822, rel=0.1)
    assert topic.MostRecentlyPublishedPapers[0]._fields == expected_fields

    assert topics_by_year.MostRecentlyPublishedPapers is None

    assert len(topics_not_by_year.MostRecentlyPublishedPapers) == approx(5862, rel=0.1)
    assert topics_not_by_year.MostRecentlyPublishedPapers[0]._fields == expected_fields
    assert topics_not_by_year.MostRecentlyPublishedPapers[0].entity_id == 3909
    assert topics_not_by_year.MostRecentlyPublishedPapers[0].entity_name == "Point Cloud Semantic Segmentation"
    assert topics_not_by_year.MostRecentlyPublishedPapers[0].publication_id == 105016011921

    assert specific_metrics.MostRecentlyPublishedPapers is None

    assert non_existing.MostRecentlyPublishedPapers is None


def test_related_topics():
    """Test RelatedTopics property"""
    expected_fields = ('entity_id', 'entity_name', 'related_topic_id', 'related_topic_name', 'related_topic_uri', 'prominencePercentile', 'relationScore', 'relatedTopicRank')

    assert len(topic.RelatedTopics) == 50
    assert topic.RelatedTopics[0]._fields == expected_fields

    assert topics_by_year.RelatedTopics is None

    assert len(topics_not_by_year.RelatedTopics) == 100
    assert topics_not_by_year.RelatedTopics[0]._fields == expected_fields
    assert topics_not_by_year.RelatedTopics[0].entity_id == 3909
    assert topics_not_by_year.RelatedTopics[0].entity_name == "Point Cloud Semantic Segmentation"
    assert topics_not_by_year.RelatedTopics[0].related_topic_id == 10013
    assert topics_not_by_year.RelatedTopics[0].related_topic_name == "Sensor Fusion and Object Detection in Autonomous Driving"
    assert topics_not_by_year.RelatedTopics[0].related_topic_uri == "Topic/10013"
    assert topics_not_by_year.RelatedTopics[0].prominencePercentile == 0
    assert topics_not_by_year.RelatedTopics[0].relationScore == 0
    assert topics_not_by_year.RelatedTopics[0].relatedTopicRank == 0

    assert specific_metrics.RelatedTopics is None

    assert non_existing.RelatedTopics is None


def test_scholarly_output():
    """Test ScholarlyOutput property"""
    expected_fields = ("entity_id", "entity_name", "metric", "year", "value", "percentage", "threshold")

    assert len(topic.ScholarlyOutput) == 1
    assert topic.ScholarlyOutput[0]._fields == expected_fields

    assert len(topics_by_year.ScholarlyOutput) == 15
    assert topics_by_year.ScholarlyOutput[0]._fields == expected_fields

    assert len(topics_not_by_year.ScholarlyOutput) == 2
    assert topics_not_by_year.ScholarlyOutput[0]._fields == expected_fields
    assert topics_not_by_year.ScholarlyOutput[0].entity_id == 3909
    assert topics_not_by_year.ScholarlyOutput[0].entity_name == "Point Cloud Semantic Segmentation"
    assert topics_not_by_year.ScholarlyOutput[0].metric == "ScholarlyOutput"
    assert topics_not_by_year.ScholarlyOutput[0].year == "all"
    assert topics_not_by_year.ScholarlyOutput[0].value == approx(5275, rel=0.1)
    assert topics_not_by_year.ScholarlyOutput[0].percentage is None
    assert topics_not_by_year.ScholarlyOutput[0].threshold is None

    assert specific_metrics.ScholarlyOutput is None

    assert non_existing.ScholarlyOutput is None


def test_top_authors():
    """Test TopAuthors property"""
    expected_fields = ('entity_id', 'entity_name', 'author_id', 'author_name', 'publicationCount')

    assert len(topic.TopAuthors) == 100
    assert topic.TopAuthors[0]._fields == expected_fields
    assert topic.TopAuthors[0].entity_id == 9549
    assert topic.TopAuthors[0].entity_name == "Quantum States and Entanglement in Many-Body Systems"
    assert topic.TopAuthors[0].author_id == 24773004400
    assert topic.TopAuthors[0].author_name == "Cirac, J. Ignacio"
    assert topic.TopAuthors[0].publicationCount == approx(23, rel=0.1)

    assert topics_by_year.TopAuthors is None

    assert len(topics_not_by_year.TopAuthors) == 200
    assert topics_not_by_year.TopAuthors[0]._fields == expected_fields

    assert non_existing.TopAuthors is None


def test_top_cited_publications():
    """Test TopCitedPublications property"""
    expected_fields = ('entity_id', 'entity_name', 'publication_id', 'citationCount')

    assert len(topic.TopCitedPublications) == 100
    assert topic.TopCitedPublications[0]._fields == expected_fields
    assert topic.TopCitedPublications[0].entity_id == 9549
    assert topic.TopCitedPublications[0].entity_name == "Quantum States and Entanglement in Many-Body Systems"
    assert topic.TopCitedPublications[0].publication_id == 85122122728
    assert topic.TopCitedPublications[0].citationCount == approx(455, rel=0.1)

    assert len(topics_by_year.TopCitedPublications) == 300
    assert topics_by_year.TopCitedPublications[0]._fields == expected_fields

    assert len(topics_not_by_year.TopCitedPublications) == 200
    assert topics_not_by_year.TopCitedPublications[0]._fields == expected_fields

    assert non_existing.TopCitedPublications is None


def test_top_institutions():
    """Test TopInstitutions property"""
    expected_fields = ('entity_id', 'entity_name', 'institution_id', 'institution_name', 'publicationCount')

    assert len(topic.TopInstitutions) == 100
    assert topic.TopInstitutions[0]._fields == expected_fields
    assert topic.TopInstitutions[0].entity_id == 9549
    assert topic.TopInstitutions[0].entity_name == "Quantum States and Entanglement in Many-Body Systems"
    assert topic.TopInstitutions[0].institution_id == 309054
    assert topic.TopInstitutions[0].institution_name == "Technical University of Munich"
    assert topic.TopInstitutions[0].publicationCount == approx(90, rel=0.1)

    assert topics_by_year.TopInstitutions is None

    assert len(topics_not_by_year.TopInstitutions) == 200
    assert topics_not_by_year.TopInstitutions[0]._fields == expected_fields

    assert non_existing.TopInstitutions is None


def test_top_journals():
    """Test TopJournals property"""
    expected_fields = ('entity_id', 'entity_name', 'journal_id', 'journal_name', 'publicationCount', 'citationCount', 'authorCount', 'publicationGrowth', 'authorGrowth', 'sjr', 'snip', 'citeScore')

    assert len(topic.TopJournals) == 100
    assert topic.TopJournals[0]._fields == expected_fields
    assert topic.TopJournals[0].entity_id == 9549
    assert topic.TopJournals[0].entity_name == "Quantum States and Entanglement in Many-Body Systems"
    assert topic.TopJournals[0].journal_id == 21100874236
    assert topic.TopJournals[0].journal_name == "Physical Review B"
    assert topic.TopJournals[0].publicationCount == approx(180, rel=0.1)
    assert topic.TopJournals[0].citationCount == approx(1597, rel=0.1)
    assert topic.TopJournals[0].authorCount == approx(396, rel=0.1)
    assert topic.TopJournals[0].publicationGrowth == approx(-37.77778, rel=0.1)
    assert topic.TopJournals[0].authorGrowth == approx(-36.49635, rel=0.1)

    assert topics_by_year.TopJournals is None

    assert len(topics_not_by_year.TopJournals) == 100
    assert topics_not_by_year.TopJournals[0]._fields == expected_fields

    assert non_existing.TopJournals is None


def test_top_keywords():
    """Test TopKeywords property"""
    expected_fields = ('entity_id', 'entity_name', 'keyword_name', 'keyword_uri', 'weight', 'relevance', 'publicationCount', 'publicationGrowth')

    assert len(topic.TopKeywords) == 100
    assert topic.TopKeywords[0]._fields == expected_fields
    assert topic.TopKeywords[0].entity_id == 9549
    assert topic.TopKeywords[0].entity_name == "Quantum States and Entanglement in Many-Body Systems"
    assert topic.TopKeywords[0].keyword_name == "Quantum Entanglement"
    assert topic.TopKeywords[0].keyword_uri == "Keyword/10253440894"
    assert topic.TopKeywords[0].weight == approx(145, rel=0.1)
    assert topic.TopKeywords[0].relevance == approx(1, rel=0.1)
    assert topic.TopKeywords[0].publicationCount == approx(165, rel=0.1)
    assert topic.TopKeywords[0].publicationGrowth == approx(257.14285, rel=0.3)

    assert len(topics_not_by_year.TopKeywords) == 200
    assert topics_not_by_year.TopKeywords[0]._fields == expected_fields

    assert topics_by_year.TopKeywords is None

    assert non_existing.TopKeywords is None



"""Base class object for superclasses."""

from json import dumps, loads
from math import ceil
from time import localtime, strftime, time

from tqdm import tqdm

from pybliometrics.exception import ScopusQueryError
from pybliometrics.utils import get_content, parse_content, SEARCH_MAX_ENTRIES
from pybliometrics.utils import listify


class Base:
    def __init__(self,
                 params: dict,
                 url: str,
                 download: bool = True,
                 verbose: bool = False,
                 **kwds: str
                 ) -> None:
        """Class intended as base class for superclasses.

        :param params: Dictionary used as header during the API request.
        :param url: The URL to be accessed.
        :param download: Whether to download the query or not.  Has no effect
                         for retrieval requests.
        :param verbose: Whether to print a download progress bar.
        :param kwds: Keywords passed on `get_content()`

        Raises
        ------
        ValueError
            If `self._refresh` is neither boolean nor numeric.
        """
        api = self.__class__.__name__
        # Checks
        try:
            _ = int(self._refresh)
        except ValueError:
            msg = "Parameter refresh needs to be numeric or boolean."
            raise ValueError(msg)

        # Compare age of file to test whether we refresh
        self._refresh, mod_ts = _check_file_age(self)

        # Read or download, possibly with caching
        fname = self._cache_file_path

        # Check if search request
        search_request = "query" in params
        # Check if ref retrieval for abstract
        ab_ref_retrieval = (api == 'AbstractRetrieval') and (params['view'] == 'REF')
        # Check if object retrieval
        obj_retrieval = (api == 'ObjectRetrieval')

        if fname.exists() and not self._refresh:
            self._mdate = mod_ts
            if search_request:
                self._json = [loads(line) for line in
                              fname.read_text().split("\n") if line]
                self._n = len(self._json)
            elif obj_retrieval:
                self._object = fname.read_bytes()
            else:
                self._json = loads(fname.read_text())
        else:
            resp = get_content(url, api, params, **kwds)
            header = resp.headers

            if ab_ref_retrieval:
                kwds['startref'] = '1'
                data = _get_all_refs(url, params, verbose, resp, **kwds)
                self._json = data
                data = [data]
            elif search_request:
                # Get number of results
                res = resp.json()
                n = int(res['search-results'].get('opensearch:totalResults', 0) or 0)
                self._n = n
                # Results size check
                cursor_exists = "cursor" in params
                if not cursor_exists and n > SEARCH_MAX_ENTRIES:
                    # Stop if there are too many results
                    text = f'Found {n:,} matches.  The query fails to return '\
                           f'more than {SEARCH_MAX_ENTRIES} entries.  Change '\
                           'your query such that it returns fewer entries.'
                    raise ScopusQueryError(text)
                self._json = []
                # Download results page-wise
                if download:
                    data = res.get('search-results', {}).get('entry', [])
                    if not n:
                        data = ""
                    if not cursor_exists:
                        start = params["start"]
                    # Download the remaining information in chunks
                    if verbose:
                        print(f'Downloading results for query "{params["query"]}":')
                    n_chunks = ceil(n/params['count'])
                    for i in tqdm(range(1, n_chunks), disable=not verbose,
                                  initial=1, total=n_chunks):
                        if cursor_exists:
                            cursor = res['search-results']['cursor']['@next']
                            params.update({'cursor': cursor})
                        else:
                            start += params["count"]
                            params.update({'start': start})
                        resp = get_content(url, api, params, **kwds)
                        res = resp.json()
                        data.extend(res.get('search-results', {}).get('entry', []))
                    header = resp.headers  # Use header of final call
                    self._json = data
                else:
                    data = None
            elif obj_retrieval:
                self._object = resp.content
                data = []
            else:
                data = loads(resp.text)
                self._json = data
                data = [data]
            # Set private variables
            self._mdate = time()
            self._header = header
            # Finally write data unless download=False
            if download:
                if obj_retrieval:
                    fname.write_bytes(self._object)
                else:
                    text = [dumps(item, separators=(',', ':')) for item in data]
                    fname.write_text("\n".join(text))

    def get_cache_file_age(self) -> int:
        """Return the age of the cached file in days."""
        diff = time() - self._mdate
        return int(diff / 86400)

    def get_cache_file_mdate(self) -> str:
        """Return the modification date of the cached file."""
        return strftime('%Y-%m-%d %H:%M:%S', localtime(self._mdate))

    def get_key_remaining_quota(self) -> str | None:
        """Return number of remaining requests for the current key and the
        current API (relative on last actual request).
        """
        try:
            return self._header['X-RateLimit-Remaining']
        except AttributeError:
            return None

    def get_key_reset_time(self) -> str | None:
        """Return time when current key is reset (relative on last
        actual request).
        """
        try:
            date = int(self._header['X-RateLimit-Reset'])
            return strftime('%Y-%m-%d %H:%M:%S', localtime(date))
        except AttributeError:
            return None


def _check_file_age(self):
    """Whether a file needs to be refreshed based on its age."""
    refresh = self._refresh
    try:
        mod_ts = self._cache_file_path.stat().st_mtime
        if not isinstance(refresh, bool):
            diff = time() - mod_ts
            days = int(diff / 86400) + 1
            allowed_age = int(self._refresh)
            refresh = allowed_age < days
    except FileNotFoundError:
        refresh = True
        mod_ts = None
    return refresh, mod_ts


def _get_all_refs(url: str, params: dict, verbose: bool, resp: dict, **kwds) -> dict:
    """Get all references for `AbstractRetrieval` with view `REF`."""
    # startref starts at 1 (0 does not work)
    # Max refs per query are 40
    # Use of refcount leads to errors
    res = resp.json()
    path_total_references = ['abstracts-retrieval-response', 'references', '@total-references']
    try:
        n = int(parse_content.chained_get(res, path_total_references))
    except TypeError:
        return res

    data = res  # data is used to gather all responses. res is a tmp variable
    path_reference = ['abstracts-retrieval-response', 'references', 'reference']
    ref_len = len(parse_content.chained_get(data, path_reference))
    n_chunks = ceil(n/ref_len)

    for i in tqdm(range(1, n_chunks), disable=not verbose,
                  initial=1, total=n_chunks):
        # Increment startref
        kwds['startref'] = str(int(kwds['startref']) + ref_len)
        # Get
        resp = get_content(url, 'AbstractRetrieval', params, **kwds)
        res = resp.json()
        res = parse_content.chained_get(res, path_reference)
        # Append
        data['abstracts-retrieval-response']['references']['reference'].extend(listify(res))
        if verbose:
            print(f'Extracted:\n\tFrom: {kwds["startref"]}\n\tTo:{len(parse_content.chained_get(data, ["abstracts-retrieval-response", "references", "reference"]))}')

    if verbose:
        print(f'Total data: {len(parse_content.chained_get(data, ["abstracts-retrieval-response", "references", "reference"]))}')

    return data



"""Test NonserialTitle()."""

from pybliometrics.sciencedirect import NonserialTitle, init

init()

nst_1 = NonserialTitle('978-0-12-823751-9', view='STANDARD', refresh=30)
nst_2 = NonserialTitle(9780128203101, view='STANDARD', refresh=30)
nst_3 = NonserialTitle('978-0-12-821777-1', view='STANDARD', refresh=30)

def test_aggregation_type():
    assert nst_1.aggregation_type == "ebook"
    assert nst_2.aggregation_type == "ebook"
    assert nst_3.aggregation_type == "ebook"


def test_authors():
    assert nst_1.authors == "Ian Newton"
    assert nst_2.authors is None
    assert nst_3.authors is None


def test_description():
    assert nst_1.description is None
    assert nst_2.description is None
    assert nst_3.description is None


def test_edition():
    assert nst_1.edition == "Second Edition"
    assert nst_2.edition is None
    assert nst_3.edition is None


def test_editors():
    assert nst_1.editors is None
    assert nst_2.editors == "Elias Barriga and Ivana Pajic-Lijakovic"
    assert nst_3.editors == "Fatos Xhafa, Mohamed A. Tawhid, Pardeep Kumar and Yugal Kumar"


def test_isbn():
    assert nst_1.isbn == "9780128237519"
    assert nst_2.isbn == "9780128203101"
    assert nst_3.isbn == "9780128217771"


def test_link_coverimage():
    assert nst_1.link_coverimage == "https://api.elsevier.com/content/nonserial/title/isbn/9780128237519?view=coverimage"
    assert nst_2.link_coverimage == "https://api.elsevier.com/content/nonserial/title/isbn/9780128203101?view=coverimage"
    assert nst_3.link_coverimage == "https://api.elsevier.com/content/nonserial/title/isbn/9780128217771?view=coverimage"


def test_link_homepage():
    assert nst_1.link_homepage == "https://www.sciencedirect.com/science/book/9780128237519"
    assert nst_2.link_homepage == "https://www.sciencedirect.com/science/book/9780128203101"
    assert nst_3.link_homepage == "https://www.sciencedirect.com/science/book/9780128217771"


def test_link_search():
    assert nst_1.link_search == "https://api.elsevier.com/content/nonserial/title/isbn/9780128237519"
    assert nst_2.link_search == "https://api.elsevier.com/content/nonserial/title/isbn/9780128203101"
    assert nst_3.link_search == "https://api.elsevier.com/content/nonserial/title/isbn/9780128217771"


def test_publisher_id():
    assert nst_1.publisher_id == "350"
    assert nst_2.publisher_id == "350"
    assert nst_3.publisher_id == "350"


def test_publisher_name():
    assert nst_1.publisher_name == "Academic Press"
    assert nst_2.publisher_name == "Academic Press"
    assert nst_3.publisher_name == "Academic Press"


def test_self_link():
    assert nst_1.self_link == "https://api.elsevier.com/content/nonserial/title/isbn/9780128237519"
    assert nst_2.self_link == "https://api.elsevier.com/content/nonserial/title/isbn/9780128203101"
    assert nst_3.self_link == "https://api.elsevier.com/content/nonserial/title/isbn/9780128217771"

def test_str():
    expected_1 = "Ian Newton. The Migration Ecology of Birds, Second Edition. Academic Press. ISBN: 9780128237519"
    expected_2 = "Elias Barriga and Ivana Pajic-Lijakovic (ed.). Viscoelasticity and  Collective Cell Migration. Academic Press. ISBN: 9780128203101"
    expected_3 = "Fatos Xhafa, Mohamed A. Tawhid, Pardeep Kumar and Yugal Kumar (ed.). Machine Learning, Big Data, and IoT for Medical Informatics. Academic Press. ISBN: 9780128217771"
    assert str(nst_1) == expected_1
    assert str(nst_2) == expected_2
    assert str(nst_3) == expected_3


def test_title():
    assert nst_1.title == "The Migration Ecology of Birds"
    assert nst_2.title == "Viscoelasticity and  Collective Cell Migration"
    assert nst_3.title == "Machine Learning, Big Data, and IoT for Medical Informatics"



from typing import NamedTuple
import warnings

from pybliometrics.superclasses import Retrieval
from pybliometrics.utils import chained_get, check_parameter_value, \
    get_link, make_float_if_possible, make_int_if_possible, VIEWS


class Citescoreinfolist(NamedTuple):
    year: int
    citescore: float
    status: str | None = None
    documentcount: int | None = None 
    citationcount: int | None = None
    percentcited: int | None = None
    rank: list['Citescoresubjectrank'] | None = None


class Citescoresubjectrank(NamedTuple):
    subjectcode: int
    rank: int
    percentile: int


class Subjectarea(NamedTuple):
    area: str
    abbreviation: str
    code: int


class Yearlydata(NamedTuple):
    year: int
    publicationcount: int
    revpercent: float | None
    zerocitessce: int
    zerocitespercentsce: float | None
    citecountsce: int


class SerialTitleISSN(Retrieval):
    @property
    def aggregation_type(self) -> str:
        """The type of the source."""
        return self._entry['prism:aggregationType']

    @property
    def citescoreyearinfolist(self) -> list[Citescoreinfolist] | None:
        """A list of named tuples of the form: `(year citescore)` or (when
        `view="CITESCORE"`) `(year citescore status documentcount citationcount
        percentcited rank)`.  `rank` is `None` or a named tuple of the form
        `(subjectcode rank percentile)`.

        For more information see the
        [CiteScore documentation](https://service.elsevier.com/app/answers/detail/a_id/14880/supporthub/scopus/).
        """
        
        try:
            data = self._entry.get('citeScoreYearInfoList', {})
        except KeyError:
            return None

        # Function to create a named tuple for CurrentMetric and CurrentTracker
        def _create_namedtuple(data, mode, info_class):
            year = data.get(f'citeScore{mode}Year')
            cite_score = data.get(f'citeScore{mode}')

            # To be consistent with old version
            if (year is None) and (cite_score is None):
                return None

            return info_class(year=int(year), citescore=float(cite_score))

        # Extract depending on view
        new_data = None
        if self._view in ('STANDARD', 'ENHANCED'):
            current = _create_namedtuple(data, 'CurrentMetric', Citescoreinfolist)
            tracker = _create_namedtuple(data, 'Tracker', Citescoreinfolist)
            new_data = [current, tracker]

        elif self._view == 'CITESCORE':
            new_data = _get_all_cite_score_years(Citescoreinfolist, Citescoresubjectrank, data)

        return new_data or None

    @property
    def eissn(self) -> str | None:
        """The electronic ISSN of the source."""
        return self._entry.get('prism:eIssn')

    @property
    def issn(self) -> str | None:
        """The ISSN of the source."""
        return self._entry.get('prism:issn')

    @property
    def oaallowsauthorpaid(self) -> str | None:
        """Whether under the Open-Access policy authors are allowed to pay."""
        return self._entry.get('oaAllowsAuthorPaid')

    @property
    def openaccess(self) -> int | None:
        """Open Access status (0 or 1)."""
        return make_int_if_possible(chained_get(self._entry, ['openaccess']))

    @property
    def openaccessstartdate(self):
        """Starting availability date."""
        return self._entry.get('openaccessStartDate')

    @property
    def openaccesstype(self) -> str | None:
        """Open Archive status (full or partial)."""
        return self._entry.get('openaccessType')

    @property
    def openaccessarticle(self) -> bool | None:
        """Open Access status."""
        return self._entry.get('openaccessArticle')

    @property
    def openarchivearticle(self) -> bool | None:
        """Open Archive status."""
        return self._entry.get('openArchiveArticle')

    @property
    def openaccesssponsorname(self) -> str | None:
        """The name of the Open Access sponsor."""
        return self._entry.get('openaccessSponsorName')

    @property
    def openaccesssponsortype(self) -> str | None:
        """The type of the Open Access sponsor."""
        return self._entry.get('openaccessSponsorType')

    @property
    def openaccessuserlicense(self) -> str | None:
        """The User license."""
        return self._entry.get('openaccessUserLicense')

    @property
    def publisher(self) -> str:
        """The publisher of the source."""
        return self._entry['dc:publisher']

    @property
    def scopus_source_link(self) -> str:
        """URL to info site on scopus.com."""
        return get_link(self._entry, 0, ["link"])

    @property
    def self_link(self) -> str:
        """URL to the source's API page."""
        return get_link(self._json, 0, ["link"])

    @property
    def sjrlist(self) -> list[tuple[int, float]] | None:
        """The SCImago Journal Rank (SJR) indicator as list of tuples in the form
        `(year, indicator)`.  See
        https://www.scimagojr.com/journalrank.php.
        """
        return _parse_list(self._entry, "SJR")

    @property
    def sniplist(self) -> list[tuple[int, float]] | None:
        """The Source-Normalized Impact per Paper (SNIP) as list of tuples in the form
        `(year, indicator)`.  See
        https://blog.scopus.com/posts/journal-metrics-in-scopus-source-normalized-impact-per-paper-snip.
        """
        return _parse_list(self._entry, "SNIP")

    @property
    def source_id(self) -> int:
        """The Scopus ID of the source."""
        return int(self._entry['source-id'])

    @property
    def subject_area(self) -> list[Subjectarea] | None:
        """List of named tuples of subject areas in the form
        `(area, abbreviation, code)` of the source.
        """
        areas = [Subjectarea(area=item['$'], code=int(item['@code']),
                      abbreviation=item['@abbrev'])
                 for item in self._entry["subject-area"]]
        return areas or None

    @property
    def title(self) -> str:
        """The title of the source."""
        return self._entry['dc:title']

    @property
    def yearly_data(self) -> list[Yearlydata] | None:
        """Yearly citation information as a list of Yearlydata in the
        form `(year, publicationcount, revpercent, zerocitessce,
        zerocitespercentsce, citecountsce)`.  That's the number of documents
        published in this year, the share of review articles thereof, the
        number and share of not-cited documents, and the number of distinct
        documents that were cited in this year.
        """
        try:
            data = self._entry['yearly-data']["info"]
        except KeyError:
            return None
        data = [Yearlydata(year=int(d['@year']), citecountsce=int(d['citeCountSCE']),
                    publicationcount=int(d['publicationCount']),
                    revpercent=make_float_if_possible(d['revPercent']),
                    zerocitessce=int(d['zeroCitesSCE']),
                    zerocitespercentsce=make_float_if_possible(d['zeroCitesPercentSCE']))
                for d in data]
        return data or None

    def __init__(self,
                 issn: int | str,
                 refresh: bool | int = False,
                 view: str = "ENHANCED",
                 years: str | None = None,
                 **kwds: str
                 ) -> None:
        """Interaction with the `ISSN` endpoint of the `Serial Title API`.
        Class retrieves data from both Scopus and ScienceDirect.

        :param issn: The ISSN or the E-ISSN of the source.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param view: The view of the file that should be downloaded.  Allowed
                     values: `STANDARD`, `ENHANCED`, `CITESCORE`.  For details
                     see https://dev.elsevier.com/sc_serial_title_views.html.
        :param years: A string specifying a year or range of years (combining
                      two years with a hyphen) for which yearly metric data
                      (SJR, SNIP, yearly-data) should be looked up for.  If
                      `None`, only the most recent metric data values are
                      provided. Note: If not `None`, refresh will always be `True`.
        :param kwds: Keywords passed on as query parameters.  Must contain
                     fields and values mentioned in the API specification at
                     https://dev.elsevier.com/documentation/SerialTitleAPI.wadl.

        Raises
        ------
        ValueError
            If any of the parameters `refresh` or `view` is not
            one of the allowed values.

        Notes
        -----
        The directory for cached results is `{path}/{view}/{source_id}`,
        where `path` is specified in your configuration file.
        """
        # Checks
        check_parameter_value(view, VIEWS['SerialTitleISSN'], "view")
        self._view = view

        # Force refresh when years is specified
        if years:
            refresh = True
        self._refresh = refresh

        # Load json
        self._id = str(issn)
        self._years = years
        Retrieval.__init__(self, identifier=self._id, date=years, **kwds)

        # Parse json
        self._json = self._json['serial-metadata-response']
        self._entry = self._json['entry'][0]

    def __str__(self):
        """Print a summary string."""
        date = self.get_cache_file_mdate().split()[0]
        areas = [e.area for e in self.subject_area or []]
        if len(areas) == 1:
            areas = areas[0]
        else:
            areas = " and ".join([", ".join(areas[:-1]), areas[-1]])
        s = f"'{self.title}', {self.aggregation_type} published by "\
            f"'{self.publisher}', is active in {areas}\n"
        metrics = []
        if self.sjrlist:
            metrics.append("SJR:  year value")
            for rec in self.sjrlist:
                metrics.append(f"      {rec[0]} {rec[1]}")
        if self.sniplist:
            metrics.append("SNIP: year value")
            for rec in self.sniplist:
                metrics.append(f"      {rec[0]} {rec[1]}")
        if metrics:
            s += f"Metrics as of {date}:\n    " + "\n    ".join(metrics) + "\n"
        s += f"    ISSN: {self.issn or '-'}, E-ISSN: {self.eissn or '-'}, "\
             f"Scopus ID: {self.source_id}"
        return s


def _parse_list(d, metric):
    """Auxiliary function to parse SNIP and SJR lists."""
    try:
        values = [(int(r['@year']), float(r['$'])) for r
                  in d[metric + "List"][metric]]
        return sorted(set(values))
    except (KeyError, TypeError):
        return None


def _get_all_cite_score_years(
        named_info_list: type[Citescoreinfolist],
        named_rank_list: type[Citescoresubjectrank],
        data: dict
) -> list[Citescoreinfolist] | None:
    """Auxiliary function to get all information contained in cite score 
    information list for the `CITESCORE` view.
    """
    data = data.get('citeScoreYearInfo', [])

    new_data = []
    # Iterate through years
    for d in data:
        citeScoreInfo = d.get('citeScoreInformationList', [])[0].get('citeScoreInfo', [])[0]
        # Iterate through subject ranks
        subject_rank = [named_rank_list(subjectcode=int(subject['subjectCode']),
                                        rank=int(subject['rank']),
                                        percentile=int(subject['percentile']))
                        for subject in citeScoreInfo.get('citeScoreSubjectRank', [])]
        # Create named tuple with info
        Citescoreinfolist_year = named_info_list(year=int(d['@year']),
            citescore=make_float_if_possible(citeScoreInfo['citeScore']),
            status=d['@status'],
            documentcount=int(citeScoreInfo['scholarlyOutput']),
            citationcount=int(citeScoreInfo['citationCount']),
            percentcited=int(citeScoreInfo['percentCited']),
            rank=subject_rank)
        # Append new data
        new_data.append(Citescoreinfolist_year)

    return new_data or None

def SerialTitle(*args, **kwds):
    """Deprecated: Use SerialTitleISSN instead.
    This class is deprecated and will be removed.
    """
    warnings.warn(
        "SerialTitle is deprecated and will be removed with the next major release. "
        "Use SerialTitleISSN instead.",
        DeprecationWarning,
        stacklevel=2
    )
    return SerialTitleISSN(*args, **kwds)



"""Tests for `scopus.AffiliationRetrieval` module."""

from pybliometrics.scopus import AffiliationRetrieval, init
from pybliometrics.scopus.affiliation_retrieval import Variant

init()


light = AffiliationRetrieval('60000356', refresh=30, view="LIGHT")
standard = AffiliationRetrieval('60000356', refresh=30, view="STANDARD")
entitled = AffiliationRetrieval('60000356', refresh=30, view='ENTITLED')


def test_address():
    assert light.address == 'Private Bag X3'
    assert standard.address == 'Private Bag X3'
    assert entitled.address is None


def test_affiliation_name():
    assert light.affiliation_name == 'University of Cape Town'
    assert standard.affiliation_name == 'University of Cape Town'
    assert entitled.affiliation_name is None


def test_author_count():
    expected = 12800
    assert light.author_count >= expected
    assert standard.author_count >= expected
    assert entitled.author_count is None


def test_city():
    assert light.city == 'Cape Town'
    assert standard.city == 'Cape Town'
    assert entitled.city is None


def test_country():
    assert light.country == 'South Africa'
    assert standard.country == 'South Africa'
    assert entitled.country is None


def test_date_created():
    assert light.date_created is None
    assert standard.date_created == (2008, 2, 2)
    assert entitled.date_created is None


def test_document_count():
    expected = 0
    assert light.document_count >= expected
    assert standard.document_count >= expected
    assert entitled.document_count is None


def test_eid():
    assert light.eid == '10-s2.0-60000356'
    assert standard.eid == '10-s2.0-60000356'
    assert entitled.eid is None


def test_entitlement():
    assert standard.document_entitlement_status is None
    assert light.document_entitlement_status is None
    assert entitled.document_entitlement_status == 'ENTITLED'


def test_identifier():
    assert light.identifier == 60000356
    assert standard.identifier == 60000356
    assert entitled.identifier is None


def test_name_variants():
    assert isinstance(light.name_variants[0], Variant)
    assert isinstance(standard.name_variants[0], Variant)
    assert entitled.name_variants is None


def test_org_domain():
    assert light.org_domain is None
    assert standard.org_domain == 'uct.ac.za'
    assert entitled.org_domain is None


def test_org_type():
    assert light.org_type is None
    assert standard.org_type == 'univ'
    assert entitled.org_type is None


def test_org_URL():
    assert light.org_URL is None
    assert standard.org_URL == 'http://www.uct.ac.za'
    assert entitled.org_URL is None


def test_postal_code():
    assert light.postal_code is None
    assert standard.postal_code == '7701'
    assert entitled.postal_code is None


def test_scopus_affiliation_link():
    expected = 'https://www.scopus.com/affil/profile.uri?afid='\
               '60000356&partnerID=HzOxMe3b&origin=inward'
    assert light.scopus_affiliation_link == expected
    assert standard.scopus_affiliation_link == expected
    assert entitled.scopus_affiliation_link is None


def test_self_link():
    expected = 'https://api.elsevier.com/content/affiliation/affiliation_id/60000356'
    assert light.self_link == expected
    assert standard.self_link == expected
    assert entitled.self_link is None


def test_search_link():
    expected = 'https://api.elsevier.com/content/search/scopus?query=af-id%2860000356%29'
    assert light.search_link== expected
    assert standard.search_link== expected
    assert entitled.search_link is None


def test_state():
    assert light.state is None
    assert standard.state == 'Western Cape'
    assert entitled.state is None


def test_status():
    assert light.status is None
    assert standard.status == "update"
    assert entitled.status is None


def test_sort_name():
    assert light.sort_name is None
    assert standard.sort_name == 'Cape Town, University of'
    assert entitled.sort_name is None


def test_url():
    expected = 'https://api.elsevier.com/content/affiliation/affiliation_id/60000356'
    assert light.url == expected
    assert standard.url == expected
    assert entitled.url is None



"""Tests for `scopus.ScopusSearch` module."""

from pybliometrics.scopus import ScopusSearch, init
from pybliometrics.scopus.scopus_search import Document

init()

# Set to refresh=False because of citation count
s_au = ScopusSearch('AU-ID(24320488600)', unescape=False, refresh=30)
s_j = ScopusSearch('SOURCE-ID(22900) AND PUBYEAR IS 2010', unescape=True, refresh=30)
s_d = ScopusSearch("DOI(10.1038/s41556-022-01034-3)", unescape=False, refresh=30)
q_empty = 'SOURCE-ID(19700188323) AND PUBYEAR IS 1900'
s_empty = ScopusSearch(q_empty, unescape=False, refresh=30)


def test_get_eids_author():
    expected = ['2-s2.0-85193728453', '2-s2.0-85117005558',
                '2-s2.0-84937325266', '2-s2.0-26444452434']
    assert s_au.get_eids() == expected


def test_get_eids_journal():
    assert len(s_j.get_eids()) == 118


def test_get_results_size():
    assert s_au.get_results_size() == 4
    assert s_j.get_results_size() == 118
    assert s_empty.get_results_size() == 0


def test_results_author():
    received = s_au.results[-1]
    expected = Document(eid='2-s2.0-26444452434', doi='10.1016/0014-2921(92)90085-B',
        pii='001429219290085B', pubmed_id=None,
        title='Economists as policymakers: A round-table discussion. Introduction',
        subtype='ar', subtypeDescription='Article', creator='Draghi M.',
        afid=None, affilname=None, affiliation_city=None, affiliation_country=None,
        author_count='1', author_names='Draghi, Mario', author_ids='24320488600',
        author_afids=None, coverDate='1992-01-01', coverDisplayDate='April 1992',
        publicationName='European Economic Review', issn='00142921',
        source_id='20749', eIssn=None, aggregationType='Journal', volume='36',
        issueIdentifier='2-3', article_number=None, pageRange='307-309',
        description=None, authkeywords=None, citedby_count=0, openaccess=0,
        freetoread=None, freetoreadLabel=None, fund_acr=None,
        fund_no=None, fund_sponsor=None)
    assert int(received.citedby_count) > 0
    assert received._replace(citedby_count=0) == expected


def test_results_journal():
    received = s_j.results[104]  # Changed from 105 to 104 due to one less result
    abstract = "The aim of this paper is to empirically test the determinants "\
        "of Research Joint Ventures' (RJVs) group dynamics. We develop a "\
        "model based on learning and transaction cost theories, which "\
        "represent the benefits and costs of RJV participation, respectively. "\
        "According to our framework, firms at each period in time weigh the "\
        "benefits against the costs of being an RJV member. RJV dynamics can "\
        "then be interpreted as a consequence of this evolving trade-off over "\
        "time. We look at entry, turbulence and exit in RJVs that have been "\
        "set up under the US National Cooperative Research Act, which allows "\
        "for certain antitrust exemptions in order to stimulate firms to "\
        "co-operate in R&D. Accounting for unobserved project characteristics "\
        "and controlling for inter-RJV interactions and industry effects, the "\
        "Tobit panel regressions show the importance of group and time "\
        "features for an RJVs evolution. We further identify an average RJVs "\
        "long-term equilibrium size and assess its determining factors. Ours "\
        "is a first attempt to produce robust stylized facts about "\
        "co-operational short- and long-term dynamics, a neglected dimension "\
        "in research co-operations, but an important element in understanding "\
        "how collaborative learning works. © 2010 Elsevier B.V. All rights "\
        "reserved."
    keywords = 'Group processes | Learning | Panel data | Research alliance dynamics | Transaction costs'
    title = 'Learning dynamics in research alliances: A panel data analysis'
    expected = Document(eid='2-s2.0-79952579400', doi='10.1016/j.respol.2010.03.002',
        pii='S0048733310000752', pubmed_id=None, title=title, subtype='ar',
        subtypeDescription='Article', creator='Duso T.', afid='60002483;60000762;60022265',
        affilname='Universiteit van Amsterdam;Humboldt-Universität zu Berlin;Erasmus Universiteit Rotterdam',
        affiliation_city='Amsterdam;Berlin;Rotterdam', affiliation_country='Netherlands;Germany;Netherlands',
        author_count='3', author_names='Duso, Tomaso;Pennings, Enrico;Seldeslachts, Jo',
        author_ids='24281174200;56248433100;25226239100', author_afids='60000762;60022265;60002483',
        coverDate='2010-01-01', coverDisplayDate='July 2010',
        publicationName='Research Policy', issn='00487333', source_id='22900',
        eIssn=None, aggregationType='Journal', volume='39', issueIdentifier='6',
        article_number=None, pageRange='776-789', description=abstract,
        authkeywords=keywords, citedby_count=0, openaccess=0, freetoread=None,
        freetoreadLabel=None, fund_acr='DFG', fund_no='HPRN-CT-2002-00224',
        fund_sponsor='Deutsche Forschungsgemeinschaft')
    assert int(received.citedby_count) >= 1
    assert received._replace(citedby_count=0) == expected


def test_results_empty():
    assert s_empty.results is None


def test_results_unescape():
    assert s_d.results[0].afid.count(";") == 14
    assert '&' in s_d.results[0].affilname



import sys
import os
from datetime import date

sys.path.append(os.path.join(os.path.abspath(os.pardir)))
autodoc_mock_imports = ["_tkinter"]

cwd = os.getcwd()
project_root = os.path.dirname(cwd)
sys.path.insert(0, os.path.abspath('source'))
sys.path.insert(0, project_root)

# General configuration
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx_autodoc_typehints',
    'sphinx_autodoc_defaultargs',
    'sphinx_copybutton',
    'sphinx_reredirects']
copybutton_prompt_text = ">>> "

source_suffix = '.rst'
master_doc = 'index'
project = 'pybliometrics'
author = 'Michael E. Rose and John Kitchin'
copyright = f"2017-{date.today().year} {author}"

import pybliometrics

version = pybliometrics.__version__

language = 'en'

exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

pygments_style = 'sphinx'
todo_include_todos = False

# Options for HTML output
html_theme = 'alabaster'
html_theme_options = {
    'github_user': 'pybliometrics-dev',
    'github_repo': 'pybliometrics',
    'github_banner': 'true',
    'github_button': 'true',
    'github_type': 'star',
}
html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'searchbox.html',
    ]
}

# Options for HTMLHelp output
html_show_sourcelink = True
htmlhelp_basename = 'pybliometricsdoc'
autoclass_content = 'both'

# Option to group members of classes
autodoc_member_order = 'bysource'

# Type hints
autodoc_typehints = "description"
napoleon_use_param = True
typehints_document_rtype = False
rst_prolog = """
.. |default| raw:: html

    <div class="default-value-section">""" + \
    ' <span class="default-value-label">Default:</span>'

# -- Options for LaTeX output ---------------------------------------------
latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

latex_documents = [
    (master_doc, 'pybliometrics.tex', 'pybliometrics Documentation',
     author, 'manual'),
]

# Options for manual page output
man_pages = [
    (master_doc, 'pybliometrics', 'pybliometrics Documentation',
     [author], 1)
]

# Options for Texinfo output
texinfo_documents = [
    (master_doc, 'pybliometrics', 'pybliometrics Documentation',
     author, 'pybliometrics', 'One line description of project.',
     'Miscellaneous'),
]

# Redirects for moved pages
redirects = {
    'reference/scopus/SerialSearch': '/reference/scopus/SerialTitleSearch.html',
    'reference/scopus/SerialTitle': '/reference/scopus/SerialTitleISSN.html'
}



from typing import NamedTuple

from pybliometrics.superclasses import Retrieval
from pybliometrics.utils import (
    chained_get,
    check_parameter_value,
    detect_id_type,
    list_authors,
    make_bool_if_possible,
    make_int_if_possible,
    parse_pages,
    VIEWS,
)


class Author(NamedTuple):
    """Named tuple representing an author."""
    surname: str
    given_name: str


class ArticleRetrieval(Retrieval):
    @property
    def abstract(self) -> str | None:
        """The abstract of a document."""
        abstract = chained_get(self._json, ['coredata', 'dc:description'])
        if abstract:
            abstract = abstract.strip()
        return abstract

    @property
    def aggregationType(self) -> str | None:
        """The aggregation type of a document."""
        return chained_get(self._json, ['coredata', 'prism:aggregationType'])

    @property
    def authors(self) -> list[Author] | None:
        """The authors of a document."""
        out = []
        for author in chained_get(self._json, ['coredata', 'dc:creator']):
            surname, given_name = author['$'].split(', ')
            new = Author(surname=surname, given_name=given_name)
            out.append(new)
        return out or None

    @property
    def copyright(self) -> str | None:
        """The copyright of a document."""
        return chained_get(self._json, ['coredata', 'prism:copyright'])

    @property
    def coverDate(self) -> str | None:
        """The date of the cover the document is in."""
        return chained_get(self._json, ['coredata', 'prism:coverDate'])

    @property
    def coverDisplayDate(self) -> str | None:
        """The cover display date of a document."""
        return chained_get(self._json, ['coredata', 'prism:coverDisplayDate'])

    @property
    def document_entitlement_status(self) -> str | None:
        """Returns the document entitlement status, i.e. tells if the requestor 
        is entitled to the requested resource.
        Note: Only works with `ENTITLED` view.
        """
        return chained_get(self._json, ['document-entitlement', 'status'])

    @property
    def doi(self) -> str:
        """The doi of a document."""
        return chained_get(self._json, ['coredata', 'prism:doi'])

    @property
    def eid(self) -> str | None:
        """The eid of a document."""
        return chained_get(self._json, ['coredata', 'eid'])

    @property
    def endingPage(self) -> str | None:
        """The ending page of a document."""
        page = chained_get(self._json, ['coredata', 'prism:endingPage'])
        return page

    @property
    def issn(self) -> int:
        """The issn of a document."""
        return make_int_if_possible(chained_get(self._json, ['coredata', 'prism:issn']))

    @property
    def openaccess(self) -> bool:
        """The document is open access."""
        open_access = chained_get(self._json, ['coredata', 'openaccessArticle'])
        return make_bool_if_possible(open_access)

    @property
    def openaccessSponsorName(self) -> str | None:
        """The open access sponsor name of a document."""
        return chained_get(self._json, ['coredata', 'openaccessSponsorName'])

    @property
    def openaccessSponsorType(self) -> str | None:
        """The open access sponsor type of a document."""
        return chained_get(self._json, ['coredata', 'openaccessSponsorType'])

    @property
    def openaccessType(self) -> str | None:
        """The open access type of a document."""
        return chained_get(self._json, ['coredata', 'openaccessType'])

    @property
    def openaccessUserLicense(self) -> str | None:
        """The open access user license of a document."""
        return chained_get(self._json, ['coredata', 'openaccessUserLicense'])

    @property
    def openArchiveArticle(self) -> bool:
        """The document is an open archive article."""
        open_archive = chained_get(self._json, ['coredata', 'openArchiveArticle'])
        return make_bool_if_possible(open_archive)

    @property
    def originalText(self) -> str | None:
        """Complete document text."""
        return self._json.get('originalText')

    @property
    def pageRange(self) -> str | None:
        """The prism:pageRange of a document."""
        return chained_get(self._json, ['coredata', 'prism:pageRange'])

    @property
    def publicationName(self) -> str:
        """The publication name of a document (e.g. Journal of Economy and Technology)."""
        return chained_get(self._json, ['coredata', 'prism:publicationName'])

    @property
    def publisher(self) -> str | None:
        """The publisher of a document."""
        return chained_get(self._json, ['coredata', 'prism:publisher'])

    @property
    def pubType(self) -> str | None:
        """The publication type of a document."""
        return chained_get(self._json, ['coredata', 'pubType'])

    @property
    def pii(self) -> str:
        """The pii of a document."""
        return chained_get(self._json, ['coredata', 'pii'])

    @property
    def sciencedirect_link(self) -> str:
        """The ScienceDirect link of a document."""
        links = chained_get(self._json, ['coredata', 'link'])
        for link in links:
            if link['@rel'] == 'scidir':
                return link['@href']

    @property
    def self_link(self) -> str:
        """The API link of a document."""
        links = chained_get(self._json, ['coredata', 'link'])
        for link in links:
            if link['@rel'] == 'self':
                return link['@href']

    @property
    def startingPage(self) -> str | None:
        """The starting page of a document."""
        return chained_get(self._json, ['coredata', 'prism:startingPage'])

    @property
    def subjects(self) -> list[str] | None:
        """The subjects of a document."""
        subjects = chained_get(self._json, ['coredata', 'dcterms:subject'])
        return [subject['$'] for subject in subjects]

    @property
    def title(self) -> str:
        """The title of a document."""
        return chained_get(self._json, ['coredata', 'dc:title'])

    @property
    def url(self) -> str:
        """The url of a document."""
        return chained_get(self._json, ['coredata', 'prism:url'])

    @property
    def volume(self) -> int | None:
        """The prism:volume of a document."""
        vol = chained_get(self._json, ['coredata', 'prism:volume'])
        return make_int_if_possible(vol)

    def __init__(self,
                 identifier: int | str,
                 refresh: bool | int = False,
                 view: str = 'META',
                 id_type: str | None = None,
                 **kwds: str
                 ):
        """Interaction with the Article Retrieval API.
        
        :param identifier: The indentifier of an article.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param view: The view of the file that should be downloaded. Allowed values:
                     'META', 'META_ABS', 'META_ABS_REF', 'FULL', 'ENTITLED'. Default: 'META'.
        :param id_type: The type of used ID. Allowed values: `None`, 'eid', 'pii',
                        'scopus_id', 'pubmed_id' and 'doi'.  If the value is `None`,
                        pybliometrics tries to infer the ID type itself.
        """
        identifier = str(identifier)
        check_parameter_value(view, VIEWS['ArticleRetrieval'], "view")
        if id_type is None:
            id_type = detect_id_type(identifier)
        else:
            allowed_id_types = ('eid', 'pii', 'scopus_id', 'pubmed_id', 'doi')
            check_parameter_value(id_type, allowed_id_types, "id_type")

        self._view = view
        self._refresh = refresh

        Retrieval.__init__(self, identifier=identifier, id_type=id_type, **kwds)
        if self._view != "ENTITLED":
            self._json = self._json['full-text-retrieval-response']

    def __str__(self):
        s = ''
        if self._view in ('FULL', 'META_ABS', 'META'):
            if self.authors:
                if len(self.authors) > 1:
                    authors = list_authors(self.authors)
                else:
                    a = self.authors[0]
                    authors = str(a.given_name) + ' ' + str(a.surname)
            else:
                authors = "(No author found)"
            # All other information
            s += f'{authors}: "{self.title}", {self.publicationName}'
            s += f', {self.volume}' if self.volume else ''
            s += ', '
            s += parse_pages(self)
            s += f'({self.coverDate[:4]}).'
            if self.doi:
                s += f' https://doi.org/{self.doi}.'
        return s



from typing import NamedTuple

from pybliometrics.superclasses import Retrieval
from pybliometrics.utils import make_int_if_possible
from pybliometrics.utils.constants import SCIVAL_METRICS
from pybliometrics.utils.parse_metrics import extract_metric_data, extract_metric_lists, MetricData


class Topic(NamedTuple):
    """Named tuple representing a topic."""
    id: int | None
    name: str | None
    uri: str | None
    prominencePercentile: int | float | None
    scholarlyOutput: int | float | None


class CorePaper(NamedTuple):
    """Named tuple representing a core paper."""
    entity_id: int | None
    entity_name: str | None
    publication_id: int | None


class RecentPaper(NamedTuple):
    """Named tuple representing a recently published paper."""
    entity_id: int | None
    entity_name: str | None
    publication_id: int | None


class RelatedTopic(NamedTuple):
    """Named tuple representing a related topic."""
    entity_id: int | None
    entity_name: str | None
    related_topic_id: int | None
    related_topic_name: str | None
    related_topic_uri: str | None
    prominencePercentile: int | float | None
    relationScore: int | float | None
    relatedTopicRank: int | None


class TopAuthor(NamedTuple):
    """Named tuple representing a top author."""
    entity_id: int | None
    entity_name: str | None
    author_id: int | None
    author_name: str | None
    publicationCount: int | float | None


class TopPublication(NamedTuple):
    """Named tuple representing a top cited publication."""
    entity_id: int | None
    entity_name: str | None
    publication_id: int | None
    citationCount: int | float | None


class TopInstitution(NamedTuple):
    """Named tuple representing a top institution."""
    entity_id: int | None
    entity_name: str | None
    institution_id: int | None
    institution_name: str | None
    publicationCount: int | float | None


class TopJournal(NamedTuple):
    """Named tuple representing a top journal."""
    entity_id: int | None
    entity_name: str | None
    journal_id: int | None
    journal_name: str | None
    publicationCount: int | float | None
    citationCount: int | float | None
    authorCount: int | float | None
    publicationGrowth: float | None
    authorGrowth: float | None
    sjr: float | None
    snip: float | None
    citeScore: float | None


class TopKeyword(NamedTuple):
    """Named tuple representing a top keyword."""
    entity_id: int | None
    entity_name: str | None
    keyword_name: str | None
    keyword_uri: str | None
    weight: int | float | None
    relevance: int | float | None
    publicationCount: int | float | None
    publicationGrowth: float | None


class TopicLookupMetrics(Retrieval):
    @property
    def topics(self) -> list[Topic]:
        """A list of namedtuples representing topics and their basic info
        in the form `(id, name, uri, prominencePercentile, scholarlyOutput)`.
        """
        out = []

        results = self._json.get('results', [])

        for result in results:
            topic_data = result.get('topic', {})
            new = Topic(
                id=make_int_if_possible(topic_data.get('id')),
                name=topic_data.get('name'),
                uri=topic_data.get('uri'),
                prominencePercentile=topic_data.get('prominencePercentile'),
                scholarlyOutput=topic_data.get('scholarlyOutput')
            )
            out.append(new)
        return out

    @property
    def AuthorCount(self) -> list[MetricData] | None:
        """Author count metrics for each topic.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold),
        """
        return extract_metric_data(self._json, 'AuthorCount', self._by_year, "topic")

    @property
    def CitationCount(self) -> list[MetricData] | None:
        """Citation count metrics for each topic.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold),
        """
        return extract_metric_data(self._json, 'CitationCount', self._by_year, "topic")

    @property
    def CorePapers(self) -> list[CorePaper] | None:
        """Core papers for the topic.
        Returns list of CorePaper namedtuples with structure:
        (entity_id, entity_name, publication_id).
        """
        out = []
        for item in extract_metric_lists(self._json, "CorePapers", "topic"):
            entity_id = item.get('entity_id')
            entity_name = item.get('entity_name')

            for paper in item.get('values', []):
                out.append(CorePaper(
                    entity_id=entity_id,
                    entity_name=entity_name,
                    publication_id=paper.get('id')
                ))
        return out or None

    @property
    def FieldWeightedCitationImpact(self) -> list[MetricData] | None:
        """Field weighted citation impact metrics for each topic.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'FieldWeightedCitationImpact', self._by_year, "topic")

    @property
    def InstitutionCount(self) -> list[MetricData] | None:
        """Institution count metrics for each topic.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'InstitutionCount', self._by_year, "topic")

    @property
    def MostRecentlyPublishedPapers(self) -> list[RecentPaper] | None:
        """Most recently published papers for the topic.
        Returns list of RecentPaper namedtuples with structure:
        (entity_id, entity_name, publication_id).
        """
        out = []
        for item in extract_metric_lists(self._json, "MostRecentlyPublishedPapers", "topic"):
            entity_id = item.get('entity_id')
            entity_name = item.get('entity_name')

            for paper in item.get('values', []):
                out.append(RecentPaper(
                    entity_id=entity_id,
                    entity_name=entity_name,
                    publication_id=paper.get('id')
                ))
        return out or None

    @property
    def RelatedTopics(self) -> list[RelatedTopic] | None:
        """Related topics for the topic.
        Returns list of RelatedTopic namedtuples with structure:
        (entity_id, entity_name, related_topic_id, related_topic_name,
        related_topic_uri, prominencePercentile, relationScore, relatedTopicRank).
        """

        out = []
        for item in extract_metric_lists(self._json, "RelatedTopics", "topic"):
            entity_id = item.get('entity_id')
            entity_name = item.get('entity_name')

            for topic in item.get('values', []):
                topic_data = topic.get('topic', {})
                out.append(RelatedTopic(
                    entity_id=entity_id,
                    entity_name=entity_name,
                    related_topic_id=topic_data.get('id'),
                    related_topic_name=topic_data.get('name'),
                    related_topic_uri=topic_data.get('uri'),
                    prominencePercentile=topic.get('prominencePercentile'),
                    relationScore=topic.get('relationScore'),
                    relatedTopicRank=topic.get('relatedtopicrank')
                ))
        return out or None

    @property
    def ScholarlyOutput(self) -> list[MetricData] | None:
        """Scholarly output metrics for each topic.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'ScholarlyOutput', self._by_year, "topic")

    @property
    def TopAuthors(self) -> list[TopAuthor] | None:
        """Top authors for the topic.
        Returns list of TopAuthor namedtuples with structure:
        (entity_id, entity_name, author_id, author_name, publicationCount).
        """

        out = []
        for item in extract_metric_lists(self._json, "TopAuthors", "topic"):
            entity_id = item.get('entity_id')
            entity_name = item.get('entity_name')

            for author in item.get('values', []):
                out.append(TopAuthor(
                    entity_id=entity_id,
                    entity_name=entity_name,
                    author_id=author.get('id'),
                    author_name=author.get('name'),
                    publicationCount=author.get('publicationCount')
                ))
        return out or None

    @property
    def TopCitedPublications(self) -> list[TopPublication] | None:
        """Top cited publications for the topic.
        Returns list of TopPublication namedtuples with structure:
        (entity_id, entity_name, publication_id, citationCount).
        """

        out = []
        for item in extract_metric_lists(self._json, "TopCitedPublications", "topic"):
            entity_id = item.get('entity_id')
            entity_name = item.get('entity_name')

            for pub in item.get('values', []):
                out.append(TopPublication(
                    entity_id=entity_id,
                    entity_name=entity_name,
                    publication_id=pub.get('id'),
                    citationCount=pub.get('citationCount')
                ))
        return out or None

    @property
    def TopInstitutions(self) -> list[TopInstitution] | None:
        """Top institutions for the topic.
        Returns list of TopInstitution namedtuples with structure:
        (entity_id, entity_name, institution_id, institution_name, publicationCount).
        """
        out = []
        for item in extract_metric_lists(self._json, "TopInstitutions", "topic"):
            entity_id = item.get('entity_id')
            entity_name = item.get('entity_name')

            for institution in item.get('values', []):
                out.append(TopInstitution(
                    entity_id=entity_id,
                    entity_name=entity_name,
                    institution_id=institution.get('id'),
                    institution_name=institution.get('name'),
                    publicationCount=institution.get('publicationCount')
                ))
        return out or None

    @property
    def TopJournals(self) -> list[TopJournal] | None:
        """Top journals for the topic.
        Returns list of TopJournal namedtuples with structure:
        (entity_id, entity_name, journal_id, journal_name, publicationCount, 
        citationCount, authorCount, publicationGrowth, authorGrowth, sjr, snip, citeScore).
        """
        out = []
        for item in extract_metric_lists(self._json, "TopJournals", "topic"):
            entity_id = item.get('entity_id')
            entity_name = item.get('entity_name')

            for journal in item.get('values', []):
                out.append(TopJournal(
                    entity_id=entity_id,
                    entity_name=entity_name,
                    journal_id=journal.get('id'),
                    journal_name=journal.get('name'),
                    publicationCount=journal.get('publicationCount'),
                    citationCount=journal.get('citationCount'),
                    authorCount=journal.get('authorCount'),
                    publicationGrowth=journal.get('publicationGrowth'),
                    authorGrowth=journal.get('authorGrowth'),
                    sjr=journal.get('sjr'),
                    snip=journal.get('snip'),
                    citeScore=journal.get('citeScore')
                ))
        return out or None

    @property
    def TopKeywords(self) -> list[TopKeyword] | None:
        """Top keywords for the topic.
        Returns list of TopKeyword namedtuples with structure:
        (entity_id, entity_name, keyword_name, keyword_uri, weight, 
        relevance, publicationCount, publicationGrowth).
        """
        out = []
        for item in extract_metric_lists(self._json, "TopKeywords", "topic"):
            entity_id = item.get('entity_id')
            entity_name = item.get('entity_name')

            for keyword in item.get('values', []):
                out.append(TopKeyword(
                    entity_id=entity_id,
                    entity_name=entity_name,
                    keyword_name=keyword.get('name'),
                    keyword_uri=keyword.get('uri'),
                    weight=keyword.get('weight'),
                    relevance=keyword.get('relevance'),
                    publicationCount=keyword.get('publicationCount'),
                    publicationGrowth=keyword.get('publicationGrowth')
                ))
        return out or None

    def __init__(self,
                 topic_ids: str | list,
                 metric_types: str | list | None = None,
                 by_year: bool = False,
                 refresh: bool | int = False,
                 **kwds: str
                 ) -> None:
        """Interaction with the SciVal's `metrics` endpoint of the `TopicLookup API`.

        :param topic_ids: SciVal Topic ID(s). Can be a single ID or comma-separated 
                          string of IDs, or a list of IDs (e.g. `[1516, 1517]`).
        :param metric_types: Metric type(s) to retrieve. Can be a single metric 
                             or comma-separated string, or a list. Available metrics are:
                             AuthorCount, CitationCount, CorePapers, FieldWeightedCitationImpact,
                             InstitutionCount, MostRecentlyPublishedPapers, RelatedTopics,
                             ScholarlyOutput, TopAuthors, TopCitedPublications, TopInstitutions,
                             TopJournals, TopKeywords.
                             If not provided, all metrics are retrieved.
        :param by_year: Whether to retrieve metrics broken down by year.
                        Note: Some metrics are not available by year: CorePapers, 
                        MostRecentlyPublishedPapers, RelatedTopics, TopAuthors, 
                        TopInstitutions, TopJournals, TopKeywords.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param kwds: Keywords passed on as query parameters.  Must contain
                     fields and values mentioned in the 
                     https://dev.elsevier.com/documentation/SciValTopicAPI.wadl.
        """
        self._view = ''
        self._refresh = refresh
        self._by_year = by_year

        # Handle topic_ids parameter
        if isinstance(topic_ids, list):
            topic_ids = ",".join(str(t) for t in topic_ids)
        else:
            topic_ids = topic_ids.replace(" ", "")

        # Depend on by_year, set default metric_types if not provided
        if metric_types is None:
            if not by_year:
                metric_types = SCIVAL_METRICS["TopicLookupMetrics"]["byYear"] + SCIVAL_METRICS["TopicLookupMetrics"]["notByYear"]
            else:
                metric_types = SCIVAL_METRICS["TopicLookupMetrics"]["byYear"]

        if isinstance(metric_types, list):
            metric_types = ",".join(metric_types)

        # Set up parameters for the API call
        params = {
            'topicIds': topic_ids,
            'metricTypes': metric_types,
            'byYear': str(by_year).lower(),
            **kwds
        }

        Retrieval.__init__(self, **params)

    def __str__(self):
        """Return pretty text version of the topic metrics."""
        topics = self.topics or []
        topic_count = len(topics)

        if topic_count == 0:
            return "No topics found"
        else:
            s = f"TopicLookupMetrics for {topic_count} topic(s):"
            for topic in topics:
                s += f"\n- {topic.name} (ID: {topic.id})"
            return s



from pybliometrics.scival.author_metrics import AuthorMetrics
from pybliometrics.utils.startup import init

init()

# Test cases as specified
single_author_all = AuthorMetrics("6602819806", by_year=False, refresh=30)
single_author_h_index = AuthorMetrics("6602819806", metric_types=["HIndices"], by_year=False, refresh=30)
multiple_authors_all = AuthorMetrics([7201667143, 6603480302], by_year=True, refresh=30)
empty_metrics = AuthorMetrics("0000000000")


def test_academic_corporate_collaboration():
    """Test AcademicCorporateCollaboration property for all test cases."""
    result = single_author_all.AcademicCorporateCollaboration

    assert has_all_fields(result[0])
    assert single_author_h_index.AcademicCorporateCollaboration is None
    assert result[0].entity_id == 6602819806
    assert result[0].entity_name == 'Algül, Hana'
    assert result[0].metric == 'Academic-corporate collaboration'
    assert result[0].year == 'all'
    assert result[0].value >= 12
    assert result[0].percentage >= 22
    assert result[0].threshold is None

    result_multi = multiple_authors_all.AcademicCorporateCollaboration
    assert has_all_fields(result_multi[0])
    assert result_multi[0].entity_id == 7201667143
    assert result_multi[0].entity_name == 'Wolff, Klaus Dietrich'
    assert result_multi[0].metric == 'Academic-corporate collaboration'
    assert result_multi[0].year >= '2024'
    assert result_multi[0].value >= 0
    assert result_multi[0].percentage >= 0
    assert result_multi[0].threshold is None


def test_academic_corporate_collaboration_impact():
    """Test AcademicCorporateCollaborationImpact property for all test cases."""
    result = single_author_all.AcademicCorporateCollaborationImpact
    assert has_all_fields(result[0])
    assert single_author_h_index.AcademicCorporateCollaborationImpact is None

    result_multi = multiple_authors_all.AcademicCorporateCollaborationImpact
    assert has_all_fields(result_multi[0])


def test_authors():
    """Test the authors property for all test cases with actual values."""
    # Test single author with all metrics
    authors = single_author_all.authors
    assert len(authors) == 1
    assert authors[0].id == 6602819806
    assert authors[0].name == "Algül, Hana"
    assert authors[0].uri == "Author/6602819806"

    # Test single author with H-indices only
    authors_h = single_author_h_index.authors
    assert len(authors_h) == 1
    assert authors_h[0].id == 6602819806
    assert authors_h[0].name == "Algül, Hana"
    assert authors_h[0].uri == "Author/6602819806"

    # Test multiple authors with actual names and IDs
    authors_multi = multiple_authors_all.authors
    assert len(authors_multi) == 2

    # Sort by ID and test
    authors_sorted = sorted(authors_multi, key=lambda x: x.id)

    assert authors_sorted[0].id == 6603480302
    assert authors_sorted[0].name == "Vogel-Heuser, Birgit"
    assert authors_sorted[0].uri == "Author/6603480302"

    assert authors_sorted[1].id == 7201667143
    assert authors_sorted[1].name == "Wolff, Klaus Dietrich"
    assert authors_sorted[1].uri == "Author/7201667143"


def has_all_fields(metric_data):
    """Check if the metric data has all required fields."""
    required_fields = ['entity_id', 'entity_name', 'metric', 'year', 'value', 'percentage', 'threshold']
    return all(hasattr(metric_data, field) for field in required_fields)


def test_citation_count():
    """Test CitationCount property for all test cases."""
    result = single_author_all.CitationCount
    assert has_all_fields(result[0])
    assert single_author_h_index.CitationCount is None

    result_multi = multiple_authors_all.CitationCount
    assert has_all_fields(result_multi[0])


def test_citations_per_publication():
    """Test CitationsPerPublication property for all test cases."""
    result = single_author_all.CitationsPerPublication
    assert has_all_fields(result[0])
    assert single_author_h_index.CitationsPerPublication is None

    result_multi = multiple_authors_all.CitationsPerPublication
    assert has_all_fields(result_multi[0])


def test_cited_publications():
    """Test CitedPublications property for all test cases."""
    result = single_author_all.CitedPublications
    assert has_all_fields(result[0])
    assert single_author_h_index.CitedPublications is None

    result_multi = multiple_authors_all.CitedPublications
    assert has_all_fields(result_multi[0])


def test_collaboration():
    """Test Collaboration property for all test cases."""
    result = single_author_all.Collaboration
    assert has_all_fields(result[0])
    assert single_author_h_index.Collaboration is None

    result_multi = multiple_authors_all.Collaboration
    assert has_all_fields(result_multi[0])


def test_collaboration_impact():
    """Test CollaborationImpact property for all test cases."""
    result = single_author_all.CollaborationImpact
    assert has_all_fields(result[0])
    assert single_author_h_index.CollaborationImpact is None

    result_multi = multiple_authors_all.CollaborationImpact
    assert has_all_fields(result_multi[0])


def test_empty_metrics():
    """Test handling of empty metric_types."""
    assert empty_metrics.authors is None 
    assert empty_metrics.CitationCount is None
    assert empty_metrics.CitationsPerPublication is None
    assert empty_metrics.CitedPublications is None
    assert empty_metrics.Collaboration is None
    assert empty_metrics.CollaborationImpact is None


def test_field_weighted_citation_impact():
    """Test FieldWeightedCitationImpact property for all test cases."""
    result = single_author_all.FieldWeightedCitationImpact
    assert has_all_fields(result[0])
    assert single_author_h_index.FieldWeightedCitationImpact is None

    result_multi = multiple_authors_all.FieldWeightedCitationImpact
    assert has_all_fields(result_multi[0])


def test_h_indices():
    """Test HIndices property for all test cases."""
    result = single_author_all.HIndices
    assert has_all_fields(result[0])

    result_h = single_author_h_index.HIndices
    assert has_all_fields(result_h[0])

    # HIndices are not available by year
    assert multiple_authors_all.HIndices is None


def test_outputs_in_top_citation_percentiles():
    """Test OutputsInTopCitationPercentiles property for all test cases."""
    result = single_author_all.OutputsInTopCitationPercentiles
    assert has_all_fields(result[0])
    assert single_author_h_index.OutputsInTopCitationPercentiles is None

    result_multi = multiple_authors_all.OutputsInTopCitationPercentiles
    assert has_all_fields(result_multi[0])


def test_publications_in_top_journal_percentiles():
    """Test PublicationsInTopJournalPercentiles property for all test cases."""
    result = single_author_all.PublicationsInTopJournalPercentiles
    assert has_all_fields(result[0])
    assert single_author_h_index.PublicationsInTopJournalPercentiles is None

    result_multi = multiple_authors_all.PublicationsInTopJournalPercentiles
    assert has_all_fields(result_multi[0])


def test_scholarly_output():
    """Test ScholarlyOutput property for all test cases."""
    result = single_author_all.ScholarlyOutput
    assert has_all_fields(result[0])
    assert single_author_h_index.ScholarlyOutput is None

    result_multi = multiple_authors_all.ScholarlyOutput
    assert has_all_fields(result_multi[0])


def test_str_representation():
    """Test the string representation of AuthorMetrics objects using actual results."""
    # Test single author with all metrics
    str_single = str(single_author_all)
    expected_single = "AuthorMetrics for 1 author(s):\n- Algül, Hana (ID: 6602819806)"
    assert str_single == expected_single

    # Test single author with H-indices only
    str_h_index = str(single_author_h_index)
    expected_h_index = "AuthorMetrics for 1 author(s):\n- Algül, Hana (ID: 6602819806)"
    assert str_h_index == expected_h_index

    # Test multiple authors
    str_multiple = str(multiple_authors_all)
    expected_multiple = "AuthorMetrics for 2 author(s):\n- Wolff, Klaus Dietrich (ID: 7201667143)\n- Vogel-Heuser, Birgit (ID: 6603480302)"
    assert str_multiple == expected_multiple

    # Test empty metrics
    str_empty = str(empty_metrics)
    expected_empty = "No authors found"
    assert str_empty == expected_empty



"""Superclass to access all Scopus retrieval APIs and dump the results."""

import hashlib
from pathlib import Path

from pybliometrics.superclasses import Base
from pybliometrics.utils import APIS_NO_ID_IN_URL, APIS_WITH_ID_TYPE, get_config, URLS


class Retrieval(Base):
    def __init__(self,
                 identifier: int | str | None = None,
                 id_type: str | None = None,
                 **kwds: str
                 ) -> None:
        """Class intended as superclass to perform retrievals.

        :param identifier: The ID to look for.
        :param id_type: The type of the used ID.  Will only take effect for
                        the Abstract Retrieval API.
        :param kwds: Keywords passed on to requests header.  Must contain
                     fields and values specified in the respective
                     API specification.

        Raises
        ------
        KeyError
            If parameter `api` is not one of the allowed values.
        """
        api = self.__class__.__name__
        # Construct URL and cache file name
        url = URLS[api]
        if api in APIS_WITH_ID_TYPE:
            url += id_type + "/"
        if api == 'CitationOverview':
            stem = identifier.replace("/", "")
            if self._citation:
                stem += "-" + self._citation
            if self._date:
                stem += "-" + self._date
        # For APIs that don't use ID in URL, hash the parameters for unique cache filename
        elif api in APIS_NO_ID_IN_URL:
            params_str = str(sorted(kwds.items()))
            stem = hashlib.md5(params_str.encode()).hexdigest()
        else:
            url += str(identifier)
            stem = str(identifier).replace('/', '_')
        # Get cache file path
        config = get_config()
        parent = Path(config.get('Directories', api))
        self._cache_file_path = parent/self._view/stem

        # Parse file contents
        params = {'view': self._view, **kwds}
        Base.__init__(self, params=params, url=url)



from pybliometrics.utils import *

from pybliometrics.sciencedirect.article_entitlement import *
from pybliometrics.sciencedirect.article_metadata import *
from pybliometrics.sciencedirect.article_retrieval import *
from pybliometrics.sciencedirect.nonserial_title import *
from pybliometrics.sciencedirect.object_metadata import *
from pybliometrics.sciencedirect.object_retrieval import *
from pybliometrics.sciencedirect.sciencedirect_search import *
from pybliometrics.sciencedirect.subject_classifications import *
from pybliometrics.scopus.serial_title_issn import *



from typing import NamedTuple

from pybliometrics.superclasses import Retrieval
from pybliometrics.utils import make_int_if_possible, chained_get


class Author(NamedTuple):
    """Named tuple representing an author."""
    id: int | None
    name: str | None
    uri: str | None


class Institution(NamedTuple):
    """Named tuple representing an institution."""
    id: int | None
    name: str | None
    country: str | None
    country_code: str | None


class PublicationLookup(Retrieval):

    @property
    def authors(self) -> list[Author] | None:
        """A list of namedtuples representing listed authors in
        the form `(id, name, uri)`.
        """
        out = []
        for item in chained_get(self._json, ['publication', 'authors'], []):
            new = Author(id=make_int_if_possible(item['id']), name=item.get('name'),
                       uri=item.get('uri'))
            out.append(new)
        return out or None

    @property
    def citation_count(self) -> int | None:
        """Count of citations."""
        return make_int_if_possible(chained_get(self._json, ['publication', 'citationCount']))

    @property
    def doi(self) -> str | None:
        """Digital Object Identifier (DOI)."""
        return chained_get(self._json, ['publication', 'doi'])

    @property
    def id(self) -> int | None:
        """ID of the document (same as EID without "2-s2.0-")."""
        return make_int_if_possible(chained_get(self._json, ['publication', 'id']))

    @property
    def institutions(self) -> list[Institution] | None:
        """A list of namedtuples representing listed institutions in
        the form `(id, name, country, country_code)`.
        """
        out = []
        for item in chained_get(self._json, ['publication', 'institutions'], []):
            new = Institution(id=make_int_if_possible(item['id']), name=item.get('name'),
                       country=item.get('country'), country_code=item.get('countryCode'))
            out.append(new)
        return out or None

    @property
    def link(self) -> str | None:
        """URL link."""
        return chained_get(self._json, ['link', '@href'])

    @property
    def publication_year(self) -> int | None:
        """Year of publication."""
        return make_int_if_possible(chained_get(self._json, ['publication', 'publicationYear']))

    @property
    def sdgs(self) -> list[str] | None:
        """Sustainable Development Goals."""
        return chained_get(self._json, ['publication', 'sdg'])

    @property
    def source_title(self) -> str | None:
        """Title of source."""
        return chained_get(self._json, ['publication', 'sourceTitle'])

    @property
    def title(self) -> str | None:
        """Publication title."""
        return chained_get(self._json, ['publication', 'title'])

    @property
    def topic_cluster_id(self) -> int | None:
        """Topic cluster id."""
        return make_int_if_possible(chained_get(self._json, ['publication', 'topicClusterId']))

    @property
    def topic_id(self) -> int | None:
        """Topic id."""
        return make_int_if_possible(chained_get(self._json, ['publication', 'topicId']))

    @property
    def type(self) -> str | None:
        """Type of publication."""
        return chained_get(self._json, ['publication', 'type'])

    def __str__(self):
        """Return pretty text version of the document."""
        if len(self.authors) >= 1:
            author_count = len(self.authors)
            authors = f"{author_count} author(s) found"
        else:
            authors = "(No author found)"

        if len(self.institutions) >= 1:
            institution_count = len(self.institutions)
            institutions = f"{institution_count} institution(s) found"
        else:
            institutions = "(No institution found)"
        return (
            f"Document with Scopus Id {self.id or 'N/A'} received:\n"
            f"- Title: {self.title}\n"
            f"- DOI: {self.doi}\n"
            f"- Type: {self.type}\n"
            f"- Publication Year: {self.publication_year}\n"
            f"- {authors}\n"
            f"- {institutions}\n"
        )

    def __init__(self,
                 identifier: int | str | None = None,
                 refresh: bool | int = False,
                 **kwds: str
                 ) -> None:
        """Interaction with the Publication Lookup API.

        :param identifier: The Scopus ID of the publication.
        :param refresh: Whether to refresh the cached file if it exists or not.
            If int is passed, cached file will be refreshed if the
            number of days since last modification exceeds that value.
        :param kwds: Keywords passed on as query parameters.  Must contain
            fields and values mentioned in the API specification at
            https://dev.elsevier.com/documentation/SciValPublicationAPI.wadl.
        """
        self._view = ''
        self._refresh = refresh
        Retrieval.__init__(self, identifier=str(identifier), **kwds)



"""Tests for `scopus.SerialTitleSearch` module."""

from pybliometrics.scopus import SerialTitleSearch, init

init()


# Search by title
ser1 = SerialTitleSearch({'title': 'SoftwareX'}, refresh=30)
# Search by ISSN
ser2 = SerialTitleSearch({'issn': '1468-0262'}, refresh=30)
# Search by publisher
ser3 = SerialTitleSearch({'pub': 'Stellenbosch'}, refresh=30)
# Search by subject abbreviation
ser4 = SerialTitleSearch({'subj': 'COMP'}, refresh=30)
# Search by subject area code
ser5 = SerialTitleSearch({'subjCode': '2612'}, refresh=30)


def test_deprecated_class():
    from pytest import deprecated_call
    from pybliometrics.scopus import SerialSearch

    with deprecated_call():
        _ = SerialSearch({'title': 'SoftwareX'}, refresh=30)


def test_results_title():
    assert len(ser1.results) == 1
    assert ser1.results[0]['title'] == 'SoftwareX'


def test_results_issn():
    assert len(ser2.results) == 1
    assert ser2.results[0]['title'] == 'Econometrica'


def test_results_pub():
    assert len(ser3.results) == 4
    assert ser3.results[0]['title'] == 'Akroterion'


def test_results_subj():
    ser4_subj_abbs = set(i['subject_area_abbrevs'] for i in ser4.results)
    assert False not in ['COMP' in i for i in ser4_subj_abbs]


def test_results_subjcode():
    ser5_subj_codes = set(i['subject_area_codes'] for i in ser5.results)
    assert False not in ['2612' in i for i in ser5_subj_codes]



from collections import defaultdict
from typing import NamedTuple

from pybliometrics.superclasses import Retrieval
from pybliometrics.utils import chained_get, check_parameter_value, \
    deduplicate, get_id, detect_id_type, get_link, listify, \
    make_int_if_possible, parse_date_created, VIEWS


class Affiliation(NamedTuple):
    id: int | None
    name: str | None
    city: str | None
    country: str | None


class AuthorGroup(NamedTuple):
    affiliation_id: int | None
    collaboration_id: str | None
    dptid: int | None
    organization: str | None
    city: str | None
    postalcode: str | None
    addresspart: str | None
    country: str | None
    auid: int | None
    orcid: str | None
    indexed_name: str | None
    surname: str | None
    given_name: str | None


class Author(NamedTuple):
    auid: int
    indexed_name: str | None
    surname: str | None
    given_name: str | None
    affiliation: str | None


class Chemical(NamedTuple):
    source: str
    chemical_name: str
    cas_registry_number: str | None


class Contributor(NamedTuple):
    given_name: str | None
    initials: str | None
    surname: str | None
    indexed_name: str | None
    role: str | None


class Correspondence(NamedTuple):
    surname: str | None
    initials: str | None
    organization: str | None
    country: str | None
    city_group: str | None


class Funding(NamedTuple):
    agency: str | None
    agency_id: str | None
    string: str | None
    funding_id: list[str] | None
    acronym: str | None
    country: str | None


class ISSN(NamedTuple):
    print: str | None = None
    electronic: str | None = None


class Reference(NamedTuple):
    position: str | None
    id: str | None
    doi: str | None
    title: str | None
    authors: str | None
    authors_auid: str | None
    authors_affiliationid: str | None
    sourcetitle: str | None
    publicationyear: str | None
    coverDate: str | None
    volume: str | None
    issue: str | None
    first: str | None
    last: str | None
    citedbycount: str | None
    type: str | None
    text: str | None
    fulltext: str | None


class Sequencebank(NamedTuple):
    name: str
    sequence_number: str
    type: str


class Area(NamedTuple):
    area: str
    abbreviation: str
    code: int


class AbstractRetrieval(Retrieval):
    @property
    def abstract(self) -> str | None:
        """The abstract of a document.
        Note: If this is empty, try `description` property instead.
        """
        return self._head.get('abstracts')

    @property
    def affiliation(self) -> list[Affiliation] | None:
        """A list of namedtuples representing listed affiliations in
        the form `(id, name, city, country)`.
        """
        out = []
        affs = listify(self._json.get('affiliation', []))
        for item in affs:
            new = Affiliation(id=make_int_if_possible(item.get('@id')), name=item.get('affilname'),
                              city=item.get('affiliation-city'),
                              country=item.get('affiliation-country'))
            out.append(new)
        return out or None

    @property
    def aggregationType(self) -> str:
        """Aggregation type of source the document is published in."""
        return chained_get(self._json, ['coredata', 'prism:aggregationType'])

    @property
    def authkeywords(self) -> list[str] | None:
        """List of author-provided keywords of the document."""
        keywords = self._json.get('authkeywords')
        if not keywords:
            return None
        else:
            try:
                return [d['$'] for d in keywords['author-keyword']]
            except TypeError:  # Singleton keyword
                return [keywords['author-keyword']['$']]

    @property
    def authorgroup(self) -> list[AuthorGroup] | None:
        """A list of namedtuples representing the article's authors and collaborations
        organized by affiliation, in the form `(affiliation_id, collaboration_id, dptid,
        organization, city, postalcode, addresspart, country, auid, orcid,
        indexed_name, surname, given_name)`.
        If `given_name` is not present, fall back to initials.
        Note: Affiliation information might be missing or mal-assigned even
        when it looks correct in the web view.  In this case please request
        a correction.  It is generally missing for collaborations.
        """
        # Information can be one of three forms:
        # 1. A dict with one key (author) or two keys (affiliation and author)
        # 2. A list of dicts with as in 1, one for each affiliation (incl. missing)
        # 3. A list of two dicts with one key each (author and collaboration)
        # Initialization
        items = listify(self._head.get('author-group', []))
        out = []
        for item in filter(None, items):
            # Get all possible items: affiliation, author, collaboration
            aff = item.get('affiliation', {})
            authors = item.get('author', [])
            collaborations = item.get('collaboration', {})
            # Affiliation information
            aff_id = make_int_if_possible(aff.get("@afid"))
            dep_id = make_int_if_possible(aff.get("@dptid"))
            org = _get_org(aff)
            # Author information
            for author in authors:
                new = AuthorGroup(affiliation_id=aff_id,
                                  organization=org,
                                  city=aff.get('city'),
                                  dptid=dep_id,
                                  postalcode=aff.get('postal-code'),
                                  addresspart=aff.get('address-part'),
                                  country=aff.get('country'),
                                  auid=make_int_if_possible(author.get('@auid')),
                                  orcid=author.get('@orcid'),
                                  surname=author.get('ce:surname'),
                                  given_name=author.get('ce:given-name', author.get('ce:initials')),
                                  indexed_name=chained_get(author, ['preferred-name', 'ce:indexed-name']),
                                  collaboration_id=None)
                out.append(new)
            # Collaboration information
            for collaboration in filter(None, listify(collaborations)):
                new = AuthorGroup(collaboration_id=collaboration.get('@collaboration-instance-id'),
                                  indexed_name=collaboration.get('ce:indexed-name'),
                                  affiliation_id=None,
                                  organization=None,
                                  city=None,
                                  dptid=None,
                                  postalcode=None,
                                  addresspart=None,
                                  country=None,
                                  auid=None,
                                  orcid=None,
                                  surname=None,
                                  given_name=None)
                out.append(new)
        return out or None

    @property
    def authors(self) -> list[Author] | None:
        """A list of namedtuples representing the article's authors, in the
        form `(auid, indexed_name, surname, given_name, affiliation)`.  In case
        multiple affiliation IDs are given, they are joined on `";"`.
        Note: The affiliation referred to here is what Scopus' algorithm
        determined as the main affiliation.  Property `authorgroup` provides
        all affiliations.
        """
        out = []
        for item in chained_get(self._json, ['authors', 'author'], []):
            affs = [a for a in listify(item.get('affiliation')) if a] or None
            try:
                aff = ";".join([aff.get('@id') for aff in affs])
            except TypeError:
                aff = None
            new = Author(auid=int(item['@auid']), surname=item.get('ce:surname'),
                         indexed_name=item.get('ce:indexed-name'), affiliation=aff,
                         given_name=chained_get(item, ['preferred-name', 'ce:given-name']))
            out.append(new)
        return out or None

    @property
    def citedby_count(self) -> int | None:
        """Number of articles citing the document."""
        path = ['coredata', 'citedby-count']
        return make_int_if_possible(chained_get(self._json, path))

    @property
    def citedby_link(self) -> str:
        """URL to Scopus page listing citing documents."""
        return get_link(self._json, 2)

    @property
    def chemicals(self) -> list[Chemical] | None:
        """List of namedtuples representing chemical entities in the form
        `(source, chemical_name, cas_registry_number)`.  In case multiple
        numbers given, they are joined on `";"`.
        """
        path = ['enhancement', 'chemicalgroup', 'chemicals']
        items = listify(chained_get(self._head, path, []))
        out = []
        for item in items:
            for chem in listify(item['chemical']):
                number = chem.get('cas-registry-number')
                try:  # Multiple numbers given
                    num = ";".join([n['$'] for n in number])
                except TypeError:
                    num = number
                new = Chemical(source=item['@source'], cas_registry_number=num,
                               chemical_name=chem['chemical-name'])
                out.append(new)
        return out or None

    @property
    def confcode(self) -> int | None:
        """Code of the conference the document belongs to."""
        return make_int_if_possible(self._confevent.get('confcode'))

    @property
    def confdate(self) -> tuple[tuple[int, int], tuple[int, int]] | None:
        """Date range of the conference the document belongs to represented
        by two tuples in the form (YYYY, MM, DD).
        """
        dates = self._confevent.get('confdate', {})
        try:
            keys = ("startdate", "enddate")
            date_order = ("@year", "@month", "@day")
            d = (tuple(int(dates[k1][k2]) for k2 in date_order) for k1 in keys)
            return tuple(d)
        except KeyError:
            return None

    @property
    def conflocation(self) -> str | None:
        """Location of the conference the document belongs to."""
        return chained_get(self._confevent, ['conflocation', 'city-group'])

    @property
    def confname(self) -> str | None:
        """Name of the conference the document belongs to."""
        return self._confevent.get('confname')

    @property
    def confsponsor(self) -> list[str] | str | None:
        """Sponsor(s) of the conference the document belongs to."""
        path = ['confsponsors', 'confsponsor']
        sponsors = chained_get(self._confevent, path, []) or []
        if len(sponsors) == 0:
            return None
        if isinstance(sponsors, list):
            return [s['$'] for s in sponsors]
        return sponsors

    @property
    def contributor_group(self) -> list[Contributor] | None:
        """List of namedtuples representing contributors compiled by Scopus,
        in the form `(given_name, initials, surname, indexed_name, role)`.
        """
        path = ['source', 'contributor-group']
        items = listify(chained_get(self._head, path, []))
        out = []
        for item in items:
            entries = listify(item.get('contributor', {}))
            for entry in entries:
                new = Contributor(
                    given_name=entry.get('ce:given-name'),
                    initials=entry.get('ce:initials'),
                    surname=entry.get('ce:surname'),
                    indexed_name=entry.get('ce:indexed-name'),
                    role=entry.get('@role')
                )
                out.append(new)
        return out or None

    @property
    def copyright(self) -> str:
        """The copyright statement of the document."""
        path = ['item', 'bibrecord', 'item-info', 'copyright', '$']
        return chained_get(self._json, path)

    @property
    def copyright_type(self) -> str:
        """The copyright holder of the document."""
        path = ['item', 'bibrecord', 'item-info', 'copyright', '@type']
        return chained_get(self._json, path)

    @property
    def correspondence(self) -> list[Correspondence] | None:
        """List of namedtuples representing the authors to whom correspondence
        should be addressed, in the form ´(surname, initials, organization,
        country, city_group)´. Multiple organziations are joined on semicolon.
        """
        items = listify(self._head.get('correspondence', []))
        out = []
        for item in items:
            aff = item.get('affiliation', {})
            try:
                org = aff['organization']
                try:
                    org = org['$']
                except TypeError:  # Multiple names given
                    org = "; ".join([d['$'] for d in org if d])
            except KeyError:
                org = None
            new = Correspondence(surname=item.get('person', {}).get('ce:surname'),
                                 initials=item.get('person', {}).get('ce:initials'),
                                 organization=org, country=aff.get('country'),
                                 city_group=aff.get('city-group'))
            out.append(new)
        return out or None

    @property
    def coverDate(self) -> str:
        """The date of the cover the document is in."""
        return chained_get(self._json, ['coredata', 'prism:coverDate'])

    @property
    def date_created(self) -> tuple[int, int, int] | None:
        """Return the `date_created` of a record.
        """
        path = ["item", "bibrecord", "item-info", "history"]
        d = chained_get(self._json, path, {})
        try:
            return parse_date_created(d)
        except KeyError:
            return None

    @property
    def description(self) -> str | None:
        """Return the description of a record.
        Note: If this is empty, try `abstract` property instead.
        """
        return chained_get(self._json, ['coredata', 'dc:description'])

    @property
    def document_entitlement_status(self) -> str | None:
        """Returns the document entitlement status, i.e. tells if the requestor 
        is entitled to the requested resource.
        Note: Only works with `ENTITLED` view.
        """
        return chained_get(self._json, ['document-entitlement', 'status'])

    @property
    def doi(self) -> str | None:
        """DOI of the document."""
        return chained_get(self._json, ['coredata', 'prism:doi'])

    @property
    def eid(self) -> str:
        """EID of the document."""
        return chained_get(self._json, ['coredata', 'eid'])

    @property
    def endingPage(self) -> str | None:
        """Ending page. If this is empty, try `pageRange` property instead."""
        # Try coredata first, fall back to head afterwards
        ending = chained_get(self._json, ['coredata', 'prism:endingPage'])
        if not ending:
            path = ['source', 'volisspag', 'pagerange', '@last']
            ending = chained_get(self._head, path)
        return ending

    @property
    def funding(self) -> list[Funding] | None:
        """List of namedtuples parsed funding information in the form
        `(agency, agency_id, string, funding_id, acronym, country)`.
        """

        def _get_funding_id(f_dict: dict) -> list:
            funding_get = f_dict.get('xocs:funding-id', [])
            try:
                return [v['$'] for v in funding_get] or None  # multiple or empty
            except TypeError:
                return [funding_get]  # single

        path = ['item', 'xocs:meta', 'xocs:funding-list', 'xocs:funding']
        funds = listify(chained_get(self._json, path, []))
        out = []
        for item in funds:
            new = Funding(agency=item.get('xocs:funding-agency'),
                          agency_id=item.get('xocs:funding-agency-id'),
                          string=item.get('xocs:funding-agency-matched-string'),
                          funding_id=_get_funding_id(item),
                          acronym=item.get('xocs:funding-agency-acronym'),
                          country=item.get('xocs:funding-agency-country'))
            out.append(new)
        return out or None

    @property
    def funding_text(self) -> str | None:
        """The raw text from which Scopus derives funding information."""
        path = ['item', 'xocs:meta', 'xocs:funding-list', 'xocs:funding-text']
        return chained_get(self._json, path)

    @property
    def isbn(self) -> tuple[str, ...] | None:
        """ISBNs `str | None` to publicationName as tuple of variying length,
        (e.g. ISBN-10 or ISBN-13)."""
        isbns = listify(chained_get(self._head, ['source', 'isbn'], []))
        if len(isbns) == 0:
            return None
        else:
            return tuple((i['$'] for i in isbns))

    @property
    def issn(self) -> ISSN | None:
        """Namedtuple in the form `(print electronic)`.
        Note: If the source has an E-ISSN, the META view will return None.
        Use FULL view instead.
        """
        container = defaultdict(lambda: None)
        # Parse information from head (from FULL view)
        info = listify(chained_get(self._head, ['source', 'issn'], []))
        for t in info:
            try:
                container[t["@type"]] = t["$"]
            except TypeError:
                container["print"] = t
        # Parse information from coredata as fallback
        fallback = chained_get(self._json, ['coredata', 'prism:issn'])
        if fallback and len(container) < 2:
            parts = fallback.split()
            if len(parts) == 2:
                if len(container) == 1:
                    for n, o in (("electronic", "print"), ("print", "electronic")):
                        if n not in container:
                            container[n] = [p for p in parts if p != container[o]]
                else:
                    # no way to find out which is which
                    pass
            else:
                container["print"] = parts[0]
        # Finalize
        if not container:
            return None
        else:
            return ISSN(**container)

    @property
    def identifier(self) -> int:
        """ID of the document (same as EID without "2-s2.0-")."""
        return get_id(self._json)

    @property
    def idxterms(self) -> list[str] | None:
        """List of index terms (these are just one category of those
        Scopus provides in the web version)
        ."""
        try:
            terms = listify(self._json.get("idxterms", {}).get('mainterm', []))
        except AttributeError:  # idxterms is empty
            return None
        try:
            return [d['$'] for d in terms] or None
        except AttributeError:
            return None

    @property
    def issueIdentifier(self) -> str | None:
        """Number of the issue the document was published in."""
        return chained_get(self._json, ['coredata', 'prism:issueIdentifier'])

    @property
    def issuetitle(self) -> str | None:
        """Title of the issue the document was published in."""
        return chained_get(self._head, ['source', 'issuetitle'])

    @property
    def language(self) -> str | None:
        """Language of the article."""
        return chained_get(self._json, ['language', '@xml:lang'])

    @property
    def openaccess(self) -> int | None:
        """The openaccess status encoded in single digits."""
        path = ['coredata', 'openaccess']
        return make_int_if_possible(chained_get(self._json, path))

    @property
    def openaccessFlag(self) -> bool | None:
        """Whether the document is available via open access or not."""
        flag = chained_get(self._json, ['coredata', 'openaccessFlag'])
        if flag:
            flag = flag == "true"
        return flag

    @property
    def pageRange(self) -> str | None:
        """Page range.  If this is empty, try `startingPage` and
        `endingPage` properties instead.
        """
        # Try data from coredata first, fall back to head afterwards
        pages = chained_get(self._json, ['coredata', 'prism:pageRange'])
        if not pages:
            return chained_get(self._head, ['source', 'volisspag', 'pages'])
        return pages

    @property
    def pii(self) -> str | None:
        """The PII (Publisher Item Identifier) of the document."""
        return chained_get(self._json, ['coredata', 'pii'])

    @property
    def publicationName(self) -> str | None:
        """Name of source the document is published in."""
        return chained_get(self._json, ['coredata', 'prism:publicationName'])

    @property
    def publisher(self) -> str | None:
        """Name of the publisher of the document.
        Note: Information provided in the FULL view of the article might be
        more complete.
        """
        # Return information from FULL view, fall back to other views
        full = chained_get(self._head, ['source', 'publisher', 'publishername'])
        if full is None:
            return chained_get(self._json, ['coredata', 'dc:publisher'])
        else:
            return full

    @property
    def publisheraddress(self) -> str | None:
        """Name of the publisher of the document."""
        return chained_get(self._head, ['source', 'publisher', 'publisheraddress'])

    @property
    def pubmed_id(self) -> int | None:
        """The PubMed ID of the document."""
        path = ['coredata', 'pubmed-id']
        return make_int_if_possible(chained_get(self._json, path))

    @property
    def refcount(self) -> int | None:
        """Number of references of an article.
        Note: Requires either the FULL view or REF view.
        """
        try:  # REF view
            return int(self._ref['@total-references'])
        except KeyError:  # FULL view
            try:
                return int(self._ref['@refcount'])
            except KeyError:
                return None

    @property
    def references(self) -> list[Reference] | None:
        """List of namedtuples representing references listed in the document,
        in the form `(position, id, doi, title, authors, authors_auid,
        authors_affiliationid, sourcetitle, publicationyear, coverDate, volume,
        issue, first, last, citedbycount, type, text, fulltext)`.

        `position` is the number at which the reference appears in the
        document, `id` is the Scopus ID of the referenced document (EID
        without the "2-s2.0-"), `authors` is a string of the names of the
        authors in the format "Surname1, Initials1; Surname2, Initials2",
        `authors_auid` is a string of the author IDs joined on "; ",
        `authors_affiliationid` is a string of the authors' affiliation IDs
        joined on "; ", `sourcetitle` is the name of the source (e.g. the
        journal), `publicationyear` is the year of the publication as string
        (FULL view only), `coverDate` is the date of the publication as string
        (REF view only), `volume` and `issue`, are strings referring to the
        volume and issue, `first` and `last` refer to the page range,
        `citedbycount` the total number of citations of the cited item (REF
        view only), `type` describes the parsing status of the reference
        (resolved or not), `text` is information on the publication,
        `fulltext` is the text the authors used for the reference.

        Note: Requires either the FULL view or REF view.
        Might be empty even if refcount is positive.  Specific fields can
        be empty.
        The lists `authors` and `authors_auid` may contain duplicates because of
        the 1:1 pairing with the list `authors_affiliationid`.
        """
        out = []
        items = listify(self._ref.get("reference", []))
        for item in items:
            try:
                info = item.get('ref-info', item)
            except AttributeError:  # item not a dictionary
                continue
            volisspag = info.get('volisspag', {}) or {}
            if isinstance(volisspag, list):
                volisspag = volisspag[0]
            volis = volisspag.get("voliss", {})
            if isinstance(volis, list):
                volis = volis[0]
            # Parse author information
            if self._view == 'FULL':  # FULL view parsing
                auth = listify(info.get('ref-authors', {}).get('author', []))
                authors = [', '.join(filter(None, [d.get('ce:surname'),
                                                   d.get('ce:initials')]))
                           for d in auth]
                auids = None
                affids = None
                ids = listify(info['refd-itemidlist']['itemid'])
                doi = _select_by_idtype(ids, id_type='DOI')
                scopus_id = _select_by_idtype(ids, id_type='SGR')
            else:  # REF view parsing
                auth = (info.get('author-list') or {}).get('author', [])
                auth = deduplicate(auth)
                authors = [', '.join(filter(None, [d.get('ce:surname'),
                                                   d.get('ce:given-name')]))
                           for d in auth]
                auids = "; ".join(filter(None, [d.get('@auid') for d in auth]))
                affs = filter(None, [d.get('affiliation') for d in auth])
                affids = "; ".join([aff.get('@id') for aff in affs])
                doi = info.get('ce:doi')
                scopus_id = info.get('scopus-id')
            # Combine information
            new = Reference(
                position=item.get('@id'),
                id=scopus_id,
                doi=doi,
                title=info.get('ref-title', {}).get('ref-titletext', info.get('title')),
                authors="; ".join(authors),
                authors_auid=auids or None,
                authors_affiliationid=affids or None,
                sourcetitle=info.get('ref-sourcetitle', info.get('sourcetitle')),
                publicationyear=info.get('ref-publicationyear', {}).get('@first'),
                coverDate=info.get('prism:coverDate'),
                volume=volis.get('@volume'),
                issue=volis.get('@issue'),
                first=volisspag.get('pagerange', {}).get('@first'),
                last=volisspag.get('pagerange', {}).get('@last'),
                citedbycount=info.get('citedby-count'),
                type=info.get('type'),
                text=info.get('ref-text'),
                fulltext=item.get('ref-fulltext')
            )
            out.append(new)
        return out or None

    @property
    def scopus_link(self) -> str:
        """URL to the document page on Scopus."""
        return get_link(self._json, 1)

    @property
    def self_link(self) -> str:
        """URL to Scopus API page of this document."""
        return get_link(self._json, 0)

    @property
    def sequencebank(self) -> list[Sequencebank] | None:
        """List of namedtuples representing biological entities defined or
        mentioned in the text, in the form `(name, sequence_number, type)`.
        """
        path = ['enhancement', 'sequencebanks', 'sequencebank']
        items = listify(chained_get(self._head, path, []))
        out = []
        for item in items:
            numbers = listify(item['sequence-number'])
            for number in numbers:
                new = Sequencebank(name=item['@name'], sequence_number=number['$'],
                                   type=number['@type'])
                out.append(new)
        return out or None

    @property
    def source_id(self) -> int | None:
        """Scopus source ID of the document."""
        path = ['coredata', 'source-id']
        return make_int_if_possible(chained_get(self._json, path))

    @property
    def sourcetitle_abbreviation(self) -> str | None:
        """Abbreviation of the source the document is published in.
        Note: Requires the FULL view of the article.
        """
        return self._head.get('source', {}).get('sourcetitle-abbrev')

    @property
    def srctype(self) -> str | None:
        """Aggregation type of source the document is published in (short
        version of aggregationType).
        """
        return chained_get(self._json, ['coredata', 'srctype'])

    @property
    def startingPage(self) -> str | None:
        """Starting page.  If this is empty, try `pageRange` property instead."""
        # Try coredata first, fall back to bibrecord afterwards
        starting = chained_get(self._json, ['coredata', 'prism:startingPage'])
        if not starting:
            path = ['source', 'volisspag', 'pagerange', '@first']
            starting = chained_get(self._head, path)
        return starting

    @property
    def subject_areas(self) -> list[Area] | None:
        """List of namedtuples containing subject areas of the article
        in the form `(area abbreviation code)`.
        Note: Requires the FULL view of the article.
        """
        path = ['subject-areas', 'subject-area']
        out = [Area(area=item['$'], abbreviation=item['@abbrev'],
                    code=int(item['@code']))
               for item in listify(chained_get(self._json, path, []))]
        return out or None

    @property
    def subtype(self) -> str:
        """Type of the document.  Refer to the Scopus Content Coverage Guide
        for a list of possible values.  Short version of subtypedescription.
        """
        return chained_get(self._json, ['coredata', 'subtype']) or None

    @property
    def subtypedescription(self) -> str:
        """Type of the document.  Refer to the Scopus Content Coverage Guide
        for a list of possible values.  Long version of subtype.
        """
        return chained_get(self._json, ['coredata', 'subtypeDescription']) or None

    @property
    def title(self) -> str | None:
        """Title of the document."""
        return chained_get(self._json, ['coredata', 'dc:title'])

    @property
    def url(self) -> str | None:
        """URL to the API view of the document."""
        return chained_get(self._json, ['coredata', 'prism:url'])

    @property
    def volume(self) -> str | None:
        """Volume for the document."""
        return chained_get(self._json, ['coredata', 'prism:volume'])

    @property
    def website(self) -> str:
        """Website of publisher."""
        path = ['source', 'website', 'ce:e-address', '$']
        return chained_get(self._head, path)

    def __init__(self,
                 identifier: int | str | None = None,
                 refresh: bool | int = False,
                 view: str = 'META_ABS',
                 id_type: str | None = None,
                 **kwds: str
                 ) -> None:
        """Interaction with the Abstract Retrieval API.

        :param identifier: The identifier of a document.  Can be the Scopus EID
                           , the Scopus ID, the PII, the Pubmed-ID or the DOI.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param id_type: The type of used ID. Allowed values: None, 'eid', 'pii',
                        'scopus_id', 'pubmed_id', 'doi'.  If the value is None,
                        the function tries to infer the ID type itself.
        :param view: The view of the file that should be downloaded.  Allowed
                     values: META, META_ABS, REF, FULL, ENTITLED, where FULL includes all
                     information of META_ABS view and META_ABS includes all
                     information of the META view.  For details see
                     https://dev.elsevier.com/sc_abstract_retrieval_views.html.
                     Note: `ENTITLED` view only contains the `document_entitlement_status`.
        :param kwds: Keywords passed on as query parameters.  Must contain
                     fields and values listed in the API specification at
                     https://dev.elsevier.com/documentation/AbstractRetrievalAPI.wadl.

        Raises
        ------
        ValueError
            If any of the parameters `id_type`, `refresh` or `view` is not
            one of the allowed values.

        Notes
        -----
        The directory for cached results is `{path}/{view}/{identifier}`,
        where `path` is specified in your configuration file.  In case
        `identifier` is a DOI, an underscore replaces the forward slash.
        """
        # Checks
        identifier = str(identifier)
        check_parameter_value(view, VIEWS['AbstractRetrieval'], "view")
        if id_type is None:
            id_type = detect_id_type(identifier)
        else:
            allowed_id_types = ('eid', 'pii', 'scopus_id', 'pubmed_id', 'doi')
            check_parameter_value(id_type, allowed_id_types, "id_type")

        # Load json
        self._view = view
        self._refresh = refresh
        Retrieval.__init__(self, identifier=identifier, id_type=id_type, **kwds)
        if self._view in ('META', 'META_ABS', 'REF', 'FULL'):
            self._json = self._json['abstracts-retrieval-response']
        self._head = chained_get(self._json, ["item", "bibrecord", "head"], {})
        conf_path = ['source', 'additional-srcinfo', 'conferenceinfo', 'confevent']
        self._confevent = chained_get(self._head, conf_path, {})
        if self._view == "REF":
            ref_path = ["references"]
        else:
            ref_path = ['item', 'bibrecord', 'tail', 'bibliography']
        self._ref = chained_get(self._json, ref_path, {})

    def __str__(self):
        """Return pretty text version of the document.

        Assumes the document is a journal article and was loaded with
        view="META_ABS" or view="FULL".
        """
        def convert_citedbycount(entry):
            try:
                return float(entry.citedbycount) or 0
            except (ValueError, TypeError):
                return 0

        def get_date(coverDate):
            try:
                return coverDate[:4]
            except TypeError:
                return None

        if self._view in ('FULL', 'META_ABS', 'META'):
            date = self.get_cache_file_mdate().split()[0]
            # Authors
            if self.authors:
                if len(self.authors) > 1:
                    authors = _list_authors(self.authors)
                else:
                    a = self.authors[0]
                    authors = str(a.given_name) + ' ' + str(a.surname)
            else:
                authors = "(No author found)"
            # All other information
            s = f'{authors}: "{self.title}", {self.publicationName}, {self.volume}'
            if self.issueIdentifier:
                s += f'({self.issueIdentifier})'
            s += ', '
            s += _parse_pages(self)
            s += f'({self.coverDate[:4]}).'
            if self.doi:
                s += f' https://doi.org/{self.doi}.\n'
            s += f'{self.citedby_count} citation(s) as of {date}'
            if self.affiliation:
                s += "\n  Affiliation(s):\n   "
                s += '\n   '.join([aff.name for aff in self.affiliation])

        elif self._view in ('REF',):
            try:
                # Sort reference list by citationcount
                top_n = 5
                references = sorted(self.references, key=convert_citedbycount, reverse=True)

                top_references = [f'{reference.title} ({get_date(reference.coverDate)}). ' +
                                  f'EID: {reference.id}' for reference in references[:top_n]]
            except TypeError:
                top_n = 0

            s = f'A total of {self.refcount or 0} references were found. '
            if top_n:
                s += f'Top {top_n} references:\n\t'
                s += '\n\t'.join(top_references)

        return s

    def get_bibtex(self) -> str:
        """Bibliographic entry in BibTeX format.

        Raises
        ------
        ValueError
            If the item's aggregationType is not Journal.
        """
        if self.aggregationType != 'Journal':
            raise ValueError('Only Journal articles supported.')
        # Item key
        year = self.coverDate[0:4]
        first = self.title.split()[0].title()
        last = self.title.split()[-1].title()
        key = ''.join([self.authors[0].surname, year, first, last])
        # Authors
        authors = ' and '.join([f"{a.given_name} {a.surname}"
                                for a in self.authors])
        # Pages
        if self.pageRange:
            pages = self.pageRange
        elif self.startingPage:
            pages = f'{self.startingPage}-{self.endingPage}'
        else:
            pages = '-'
        # All information
        bib = f"@article{{{key},\n  author = {{{authors}}},\n  title = "\
              f"{{{{{self.title}}}}},\n  journal = {{{self.publicationName}}},"\
              f"\n  year = {{{year}}},\n  volume = {{{self.volume}}},\n  "\
              f"number = {{{self.issueIdentifier}}},\n  pages = {{{pages}}}"
        # DOI
        if self.doi:
            bib += f",\n  doi = {{{self.doi}}}"
        bib += "}"
        return bib

    def get_html(self) -> str:
        """Bibliographic entry in html format."""
        # Author links
        au_link = ('<a href="https://www.scopus.com/authid/detail.url'
                   '?origin=AuthorProfile&authorId={0}">{1}</a>')
        if len(self.authors) > 1:
            authors = ', '.join([au_link.format(a.auid, a.given_name +
                                                 ' ' + a.surname)
                                 for a in self.authors[0:-1]])
            authors += (' and ' +
                        au_link.format(self.authors[-1].auid,
                                       (str(self.authors[-1].given_name) +
                                        ' ' +
                                        str(self.authors[-1].surname))))
        else:
            a = self.authors[0]
            authors = au_link.format(a.auid, a.given_name + ' ' + a.surname)
        title = f'<a href="{self.scopus_link}">{self.title}</a>'
        if self.volume and self.issueIdentifier:
            volissue = f'<b>{self.volume}({self.issueIdentifier})</b>'
        elif self.volume:
            volissue = f'<b>{self.volume}</b>'
        else:
            volissue = 'no volume'
        jlink = '<a href="https://www.scopus.com/source/sourceInfo.url'\
                f'?sourceId={self.source_id}">{self.publicationName}</a>'
        s = f"{authors}, {title}, {jlink}, {volissue}, " +\
            f"{_parse_pages(self, unicode=True)}, ({self.coverDate[:4]})."
        if self.doi:
            s += f' <a href="https://doi.org/{self.doi}">doi:{self.doi}</a>.'
        return s

    def get_latex(self) -> str:
        """Bibliographic entry in LaTeX format."""
        if len(self.authors) > 1:
            authors = _list_authors(self.authors)
        else:
            a = self.authors
            authors = ' '.join([a.given_name, a.surname])
        if self.volume and self.issueIdentifier:
            volissue = f'\\textbf{{{self.volume}({self.issueIdentifier})}}'
        elif self.volume:
            volissue = f'\\textbf{{{self.volume}}}'
        else:
            volissue = 'no volume'
        s = f'{authors}, \\textit{{{self.title}}}, {self.publicationName}, ' +\
            f'{volissue}, {_parse_pages(self)} ({self.coverDate[:4]}).'
        if self.doi:
            s += f' \\href{{https://doi.org/{self.doi}}}{{doi:{self.doi}}}, '
        s += f'\\href{{{self.scopus_link}}}{{scopus:{self.eid}}}.'
        return s

    def get_ris(self) -> str:
        """Bibliographic entry in RIS (Research Information System Format)
        format for journal articles.

        Raises
        ------
        ValueError
            If the item's aggregationType is not Journal.
        """
        if self.aggregationType != 'Journal':
            raise ValueError('Only Journal articles supported.')
        # Basic information
        ris = (
            f"TY  - JOUR\nTI  - {self.title}\nJO  - {self.publicationName}"
            f"\nDA  - {self.coverDate}\nPY  - {self.coverDate[0:4]}\n"
        )
        # Authors
        for au in self.authors:
            ris += f'AU  - {au.indexed_name}\n'
        # DOI
        if self.doi:
            ris += f'DO  - {self.doi}\nUR  - https://doi.org/{self.doi}\n'
        # Volume
        if self.volume:
            ris += f"VL  - {self.volume}\n"
        # Issue
        if self.issueIdentifier:
            ris += f'IS  - {self.issueIdentifier}\n'
        # Pages
        if self.pageRange:
            ris += f"SP  - {self.pageRange}\n"
        ris += "ER  - \n\n"
        return ris


def _get_org(aff):
    """Auxiliary function to extract org information from affiliation
    for authorgroup.
    """
    try:
        org = aff['organization']
        if not isinstance(org, str):
            try:
                org = org['$']
            except TypeError:  # Multiple names given
                org = ', '.join([d['$'] for d in org if d])
    except KeyError:  # Author group w/o affiliation
        org = None
    return org


def _list_authors(lst):
    """Format a list of authors (Surname, Firstname and Firstname Surname)."""
    authors = ', '.join([' '.join([a.given_name, a.surname]) for a in lst[0:-1]])
    authors += ' and ' + ' '.join([lst[-1].given_name, lst[-1].surname])
    return authors


def _parse_pages(self, unicode=False):
    """Auxiliary function to parse and format page range of a document."""
    if self.pageRange:
        pages = f'pp. {self.pageRange}'
    elif self.startingPage:
        pages = f'pp. {self.startingPage}-{self.endingPage}'
    else:
        pages = '(no pages found)'
    if unicode:
        pages = f'{pages}'
    return pages


def _select_by_idtype(lst, id_type):
    """Auxiliary function to return items matching a special idtype."""
    try:
        return [d['$'] for d in lst if d['@idtype'] == id_type][0]
    except IndexError:
        return None



"""Tests for `scopus.AffiliationSearch` module."""

from pybliometrics.scopus import AffiliationSearch, init
from pybliometrics.scopus.affiliation_search import Affiliation

init()

s1 = AffiliationSearch('AF-ID(60021784)', refresh=30)
s2 = AffiliationSearch('AFFIL(Max Planck Munich)', download=False, refresh=True)


def test_affiliations():
    received1 = s1.affiliations
    assert isinstance(received1, list)
    expected = Affiliation(eid='10-s2.0-60021784', name='New York University',
        variant='', documents=0, city='New York', country='United States')
    assert received1[0].documents >= 90_000
    assert received1[0]._replace(documents=0) == expected


def test_affiliations_nodownload():
    assert s2.affiliations is None


def test_get_results_size():
    received1 = s1.get_results_size()
    assert received1 >= 1
    received2 = s2.get_results_size()
    assert received2 >= 60



"""Tests for the startup and configuration module."""

from pathlib import Path

import pytest

from pybliometrics.scopus import init
from pybliometrics.utils import get_insttokens, get_keys

CURRENT_DIR = Path(__file__).resolve().parent
TEST_CONFIG = CURRENT_DIR / 'test_config.cfg'


def test_custom_insttokens():
    """Test whether custom insttokens are correctly set."""
    init(keys=['1', '2', '3'], inst_tokens=['a', 'b'])
    assert get_insttokens() == ['a', 'b']


def test_custom_keys():
    """Test whether custom keys are correctly set."""
    init(keys=['1', '2', '3'])
    assert get_keys() == ['1', '2', '3']


def test_imports():
    """Test the import and initialization of the pybliometrics package and its submodules."""
    import pybliometrics
    pybliometrics.init()

    import pybliometrics.sciencedirect
    pybliometrics.sciencedirect.init()

    import pybliometrics.scopus
    pybliometrics.scopus.init()

    import pybliometrics.scival
    pybliometrics.scival.init()


def test_new_config():
    """Test whether a new config file is created."""
    TEST_CONFIG.unlink(missing_ok=True)

    # Create new config
    init(config_path=TEST_CONFIG, keys=['3', '4', '5'], inst_tokens=['c', 'd'])

    # Use custom keys and tokens
    assert get_keys() == ['3', '4', '5']
    assert get_insttokens() == ['c', 'd']


def test_new_test_config():
    """Test whether the new test config file is correctly read."""
    init(config_path=TEST_CONFIG, keys=['3', '4', '5'])

    # Use keys and tokens from test config
    assert get_keys() == ['3', '4', '5']
    assert get_insttokens() == []

    init(config_path=TEST_CONFIG, keys=['5', '6', '7'], inst_tokens=['e', 'f'])

    # Use custom keys and tokens
    assert get_keys() == ['5', '6', '7']
    assert get_insttokens() == ['e', 'f']


def test_error_more_tokens_than_keys():
    """Test whether an error is raised if more tokens than keys are provided."""
    with pytest.raises(ValueError):
        init(keys=['1'], inst_tokens=['a', 'b'])



"""Tests for sciencedirect.ArticleMetadata"""

from pybliometrics.exception import Scopus400Error
from pybliometrics.sciencedirect import ArticleMetadata, init
from pybliometrics.sciencedirect.article_metadata import Document

init()

am_standard = ArticleMetadata('TITLE("Bayesian Network") AND YEAR(2015)',
                              view="STANDARD",
                              refresh=30)
am_complete = ArticleMetadata('AFFIL("MIT") AND YEAR("2023") AND DOI(10.1016/B978-0-32-399851-2.00030-2)',
                              view="COMPLETE",
                              refresh=30)
am_empty = ArticleMetadata('TITLE("Not a very realistic title")', view="STANDARD", refresh=30)


def test_empty_results():
    assert am_empty.results is None
    assert am_empty._n == 0


def test_all_fields():
    expected_complete_doc = Document(
        authorKeywords="Large scale learning | Few shot learning | Meta learning",
        authors="Balaji, Yogesh",
        available_online_date="2022-09-30",
        first_author="Sankaranarayanan, Swami",
        abstract_text="Meta learning techniques have been successfully used to mitigate issues such as distribution shift in neural network training. However, the proposed approaches remain complex to train and less scalable. In this chapter, we discuss some recent approaches that have successfully demonstrated the use of meta learning based techniques in a big data setting such as Imagenet and beyond.",
        doi="10.1016/B978-0-32-399851-2.00030-2",
        title="Meta learning in the big data regime Applications to transfer learning and few shot learning",
        eid="3-s2.0-B9780323998512000302",
        link="https://www.sciencedirect.com/science/article/pii/B9780323998512000302",
        openArchiveArticle=False,
        openaccess_status="0",
        openaccessArticle=False,
        openaccessUserLicense=None,
        pii="B978-0-32-399851-2.00030-2",
        aggregationType="EBook",
        copyright="Copyright © 2023 Elsevier Inc. All rights are reserved, including those for text and data mining, AI training, and similar technologies.",
        coverDate="2023-12-31",
        coverDisplayDate="2023",
        edition=None,
        endingPage="393",
        isbn="9780323998512",
        publicationName="Meta Learning With Medical Imaging and Health Informatics Applications",
        startingPage="385",
        teaser="Meta learning techniques have been successfully used to mitigate issues such as distribution shift in neural network training. However, the proposed approaches remain complex to train and less scalable....",
        api_link="https://api.elsevier.com/content/article/pii/B9780323998512000302",
        publicationType="chp",
        vor_available_online_date="2022-09-30",
    )
    assert am_complete.results[0] == expected_complete_doc

    expected_standard_doc = Document(
        authorKeywords=None,
        authors="Zhong, Lu;Haijun, Zeng",
        available_online_date="2015-02-14",
        first_author="Kang, Chen",
        abstract_text=None,
        doi="10.1016/j.proeng.2014.12.523",
        title="Research on Probabilistic Safety Analysis Approach of Flight Control System Based on Bayesian Network",
        eid="1-s2.0-S1877705814036376",
        link="https://www.sciencedirect.com/science/article/pii/S1877705814036376",
        openArchiveArticle=False,
        openaccess_status="1",
        openaccessArticle=True,
        openaccessUserLicense="http://creativecommons.org/licenses/by-nc-nd/4.0/",
        pii="S1877-7058(14)03637-6",
        aggregationType=None,
        copyright=None,
        coverDate="2015-12-31",
        coverDisplayDate="2015",
        edition=None,
        endingPage="184",
        isbn=None,
        publicationName="Procedia Engineering",
        startingPage="180",
        teaser="Traditional probabilistic safety analysis methods are not suitable for modern flight control system with multi-state probability. In this papera Bayesian Network based probabilistic safety model is...",
        api_link="https://api.elsevier.com/content/article/pii/S1877705814036376",
        publicationType="fla",
        vor_available_online_date=None,
    )
    assert am_standard.results[0] == expected_standard_doc


def test_field_consistency():
    am_wrong_field = ArticleMetadata('TITLE("Bayesian Network") AND YEAR(2015)',
                                 integrity_fields=["notExistingField"],
                                 integrity_action="warn",
                                 view="STANDARD",
                                 refresh=30)
    try:
        am_wrong_field.results
    except ValueError:
        pass
    except Exception as e:
        raise AssertionError(f"Unexpected exception type: {type(e).__name__}")
    else:
        raise AssertionError("Expected ValueError but no exception was raised")


def test_length():
    assert len(am_standard.results) == am_standard._n
    assert len(am_complete.results) == am_complete._n


def test_string():
    part_of_str = 'Search \'AFFIL("MIT") AND YEAR("2023") AND DOI(10.1016/B978-0-32-399851-2.00030-2)\' yielded 1 document as of'
    assert part_of_str in am_complete.__str__()


def test_wrong_query():
    try:
        ArticleMetadata(
            'TITLE("Bayesian Network") AND YEAR(2015', view="STANDARD", refresh=30
        )
    except Scopus400Error:
        pass
    except Exception as e:
        raise AssertionError(f"Unexpected exception type: {type(e).__name__}")
    else:
        raise AssertionError("Expected Scopus400Error but no exception was raised")



from requests import Session
from requests.adapters import HTTPAdapter
from requests.exceptions import JSONDecodeError
from urllib3.util import Retry

from pybliometrics import __version__
from pybliometrics import exception
from pybliometrics.utils.startup import get_config, get_insttokens, get_keys, _throttling_params

# Define user agent string for HTTP requests
user_agent = 'pybliometrics-v' + __version__

errors = {400: exception.Scopus400Error, 401: exception.Scopus401Error,
          403: exception.Scopus403Error, 404: exception.Scopus404Error,
          407: exception.Scopus407Error, 413: exception.Scopus413Error, 
          414: exception.Scopus414Error, 429: exception.Scopus429Error}


def get_session() -> Session:
    """Auxiliary function to create a session"""
    config = get_config()

    _retries = config.getint("Requests", "Retries", fallback=5)
    retry = Retry(total=_retries, backoff_factor=0.1,
                  status_forcelist=[500, 501, 502, 503, 504, 524])
    adapter = HTTPAdapter(max_retries=retry)
    session = Session()
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session


def get_content(url, api, params=None, **kwds):
    """Helper function to download a file and return its content.

    Parameters
    ----------
    url : str
        The URL to be parsed.

    api : str
        The Scopus API to be accessed.

    params : dict (optional)
        Dictionary containing query parameters.  For required keys
        and accepted values see e.g.
        https://api.elsevier.com/documentation/AuthorRetrievalAPI.wadl

    **kwds : key-value parings, optional
        Keywords passed on to as query parameters.  Must contain fields
        and values specified in the respective API specification.

    Raises
    ------
    ScopusHtmlError or HTTPError
        If the status of the response is not ok.

    ValueError
        If the accept parameter is not one of the accepted values.

    Returns
    -------
    resp : byte-like object
        The content of the file, which needs to be serialized.
    """
    from random import shuffle
    from time import sleep, time

    # Get needed ressources for query
    config = get_config()

    keys = get_keys()

    # Get tokens and zip with keys
    insttokens = get_insttokens()
    insttokens = list(zip(keys, insttokens))

    # Keep keys that are not insttokens
    keys = keys[len(insttokens):]

    session = get_session()

    params = params or {}
    params.update(**kwds)
    proxies = dict(config._sections.get("Proxy", {}))
    timeout = config.getint("Requests", "Timeout", fallback=20)

    # Get keys/tokens and create header
    token_key, insttoken = None, None
    if "insttoken" in params:
        token_key = params.pop("apikey")
        insttoken = params.pop("insttoken")
    elif "apikey" in params:
        key = params.pop("apikey")
    elif insttokens:
        token_key, insttoken = insttokens.pop(0)
    else:
        key = keys.pop(0)

    header = {'Accept': 'application/json',
              'User-Agent': user_agent,
              'X-ELS-APIKey': token_key or key}

    # Eventually wait bc of throttling
    if len(_throttling_params[api]) == _throttling_params[api].maxlen:
        try:
            sleep(1 - (time() - _throttling_params[api][0]))
        except (IndexError, ValueError):
            pass

    # Use insttoken if available
    if insttoken:
        header['X-ELS-Insttoken'] = insttoken
        resp = session.get(url, headers=header, params=params, timeout=timeout)
    else:
        resp = session.get(url, headers=header, params=params, timeout=timeout, proxies=proxies)

    # If 429 try other tokens
    while (resp.status_code == 429) or (resp.status_code == 401):
        try:
            token_key, token = insttokens.pop(0) # Get and remove current key
            header['X-ELS-APIKey'] = token_key
            header['X-ELS-Insttoken'] = token
            shuffle(insttokens)
            resp = session.get(url, headers=header, params=params, timeout=timeout)
        except IndexError:  # All tokens depleted
            break

   # Remove Insttoken from header (if present)
    if 'X-ELS-Insttoken' in header:
        del header['X-ELS-Insttoken']

    # If 429 try other keys
    while (resp.status_code == 429) or (resp.status_code == 401):
        try:
            key = keys.pop(0)  # Remove current key
            header['X-ELS-APIKey'] = key
            shuffle(keys)
            resp = session.get(url, headers=header, proxies=proxies, params=params, timeout=timeout)
        except IndexError:  # All keys depleted
            break

    _throttling_params[api].append(time())

    # Eventually raise error, if possible with supplied error message
    try:
        error_type = errors[resp.status_code]
        try:
            reason = resp.json()['service-error']['status']['statusText']
        except KeyError:
            try:
                reason = resp.json()['message']
            except:
                reason = ""
        raise error_type(reason)
    except (JSONDecodeError, KeyError):
        resp.raise_for_status()
    return resp


def detect_id_type(sid):
    """Method that tries to infer the type of abstract ID.

    Parameters
    ----------
    sid : str
        The ID of an abstract on Scopus.

    Raises
    ------
    ValueError
        If the ID type cannot be inferred.

    Notes
    -----
    Scopus IDs and Pubmed IDs are sometimes hard to distinguish.  If you
    work with both types, consider specifying the ID type manually.
    """
    sid = str(sid)
    if not sid.isnumeric():
        if sid.startswith('1-s2.0-') or sid.startswith('2-s2.0-'):
            id_type = 'eid'
        elif '/' in sid or "." in sid:
            id_type = 'doi'
        elif 16 <= len(sid) <= 17:
            id_type = 'pii'
    else:
        if len(sid) < 10:
            id_type = 'pubmed_id'
        else:
            id_type = 'scopus_id'
    try:
        return id_type
    except UnboundLocalError:
        raise ValueError(f'ID type detection failed for "{sid}".')



"""Tests for `scopus.SerialTitleISSN` module."""

import datetime

from pybliometrics.scopus import SerialTitleISSN, init
from pybliometrics.scopus.serial_title_issn import Citescoreinfolist, Subjectarea, Yearlydata

init()


# SoftwareX
softwarex = SerialTitleISSN("2352-7110", refresh=30)
# OECD Economic Studies
oecd = SerialTitleISSN("0255-0822", refresh=30)
# Neural Networks
neural_networks = SerialTitleISSN('1879-2782', view='CITESCORE', refresh=30)
# Empty rank for year 2018 in JCO clinical cancer informatics
jco_cci = SerialTitleISSN('2473-4276', view='CITESCORE', refresh=30)


def test_aggregation_type():
    assert softwarex.aggregation_type == "journal"
    assert oecd.aggregation_type == "journal"
    assert jco_cci.aggregation_type == "journal"


def test_citescoreyearinfolist():
    # Test softwarex
    expected_named_tuple = [Citescoreinfolist(year=2024, citescore=4.2),
                            Citescoreinfolist(year=2025, citescore=3.4)]
    assert softwarex.citescoreyearinfolist == expected_named_tuple

    # Test oecd
    assert oecd.citescoreyearinfolist[0] is None
    assert oecd.citescoreyearinfolist[1] is None

    # Test CITESCORE view
    this_year = datetime.date.today().year
    assert neural_networks.citescoreyearinfolist[0].year == this_year
    assert isinstance(neural_networks.citescoreyearinfolist[3].citationcount, int)

    # Test empty rank
    assert jco_cci.citescoreyearinfolist[-1].rank == []


def test_deprecated_class():
    from pytest import deprecated_call
    from pybliometrics.scopus import SerialTitle

    with deprecated_call():
        _ = SerialTitle("2352-7110", refresh=30)
    

def test_eissn():
    assert softwarex.eissn == "2352-7110"
    assert oecd.eissn == "1609-7491"
    assert jco_cci.eissn == "2473-4276"


def test_issn():
    assert softwarex.issn is None
    assert oecd.issn == "0255-0822"
    assert jco_cci.issn is None


def test_oaallowsauthorpaid():
    assert softwarex.oaallowsauthorpaid is None
    assert oecd.oaallowsauthorpaid is None
    assert jco_cci.oaallowsauthorpaid is None


def test_openaccess():
    assert softwarex.openaccess == 1
    assert oecd.openaccess is None
    assert jco_cci.openaccess is None


def test_openaccessstartdate():
    assert softwarex.openaccessstartdate is None
    assert oecd.openaccessstartdate is None
    assert jco_cci.openaccessstartdate is None


def test_openaccesstype():
    assert softwarex.openaccesstype is None
    assert oecd.openaccesstype is None
    assert jco_cci.openaccesstype is None


def test_openaccessarticle():
    assert softwarex.openaccessarticle == True
    assert oecd.openaccessarticle is None
    assert jco_cci.openaccessarticle is None


def test_openarchivearticle():
    assert softwarex.openarchivearticle is None
    assert oecd.openarchivearticle is None
    assert jco_cci.openarchivearticle is None


def test_openaccesssponsorname():
    assert softwarex.openaccesssponsorname is None
    assert oecd.openaccesssponsorname is None
    assert jco_cci.openaccesssponsorname is None


def test_openaccessuserlicense():
    assert softwarex.openaccessuserlicense is None
    assert oecd.openaccessuserlicense is None
    assert jco_cci.openaccessuserlicense is None


def test_publisher():
    assert softwarex.publisher == "Elsevier B.V."
    assert oecd.publisher == "OECD"
    assert jco_cci.publisher == "Lippincott Williams and Wilkins"


def test_scopus_source_link():
    expected1 = "https://www.scopus.com/source/sourceInfo.url?sourceId=21100422153"
    assert softwarex.scopus_source_link == expected1
    expected2 = "https://www.scopus.com/source/sourceInfo.url?sourceId=24107"
    assert oecd.scopus_source_link == expected2
    expected3 = "https://www.scopus.com/source/sourceInfo.url?sourceId=21100897027"
    assert jco_cci.scopus_source_link == expected3


def test_self_link():
    expected1 = "https://api.elsevier.com/content/serial/title/issn/23527110"
    assert softwarex.self_link == expected1
    expected2 = "https://api.elsevier.com/content/serial/title/issn/02550822"
    assert oecd.self_link == expected2
    expected3 = "https://api.elsevier.com/content/serial/title/issn/24734276"
    assert jco_cci.self_link == expected3


def test_sjrlist():
    assert softwarex.sjrlist == [(2024, 0.483)]
    assert oecd.sjrlist == [(1999, 2.723)]
    assert jco_cci.sjrlist == [(2024, 1.25)]


def test_sniplist():
    assert softwarex.sniplist == [(2024, 1.095)]
    assert oecd.sniplist is None
    assert jco_cci.sniplist == [(2024, 0.896)]


def test_source_id():
    assert softwarex.source_id == 21100422153
    assert oecd.source_id == 24107
    assert jco_cci.source_id == 21100897027


def test_subject_area():
    expected1 = [
        Subjectarea(area='Computer Science Applications', abbreviation='COMP', code=1706),
        Subjectarea(area='Software', abbreviation='COMP', code=1712)
    ]
    assert softwarex.subject_area == expected1
    expected2 = [
        Subjectarea(area='Geography, Planning and Development', abbreviation='SOCI', code=3305)
    ]
    assert oecd.subject_area == expected2
    expected3 = [
        Subjectarea(area='Cancer Research', abbreviation='BIOC', code=1306),
        Subjectarea(area='Oncology', abbreviation='MEDI', code=2730),
        Subjectarea(area='Health Informatics', abbreviation='MEDI', code=2718)
    ]
    assert jco_cci.subject_area == expected3


def test_title():
    assert softwarex.title == "SoftwareX"
    assert oecd.title == "OECD Economic Studies"
    assert jco_cci.title == "JCO Clinical Cancer Informatics"


def test_yearly_data():
    assert isinstance(softwarex.yearly_data, list)
    assert len(softwarex.yearly_data) == 30
    expected1_2020 = Yearlydata(year=2020, publicationcount=164, revpercent=0.0,
        zerocitessce=6, zerocitespercentsce=3.658536585365854,
        citecountsce=2580)
    assert softwarex.yearly_data[24] == expected1_2020
    expected2_1996 = Yearlydata(year=1996, publicationcount=4, revpercent=0.0,
        zerocitessce=0, zerocitespercentsce=0, citecountsce=33)
    assert oecd.yearly_data[0] == expected2_1996
    assert jco_cci.yearly_data is None



"""Base exceptions and classes for pybliometrics.scopus."""

# Base classes
class ScopusException(Exception):
    """Base class for exceptions in pybliometrics."""


class ScopusError(ScopusException):
    """Exception for a serious error in pybliometrics."""


# Query errors
class ScopusQueryError(ScopusException):
    """Exception for problems related to Scopus queries."""


# HTML errors
class ScopusHtmlError(ScopusException):
    """Wrapper for exceptions raised by requests."""


class Scopus400Error(ScopusHtmlError):
    """Raised if a query yields a 400 error (Bad Request for url)."""


class Scopus401Error(ScopusHtmlError):
    """Raised if a query yields a 401 error (Unauthorized for url)."""


class Scopus403Error(ScopusHtmlError):
    """Raised if a query yields a 403 error (Forbidden for url)."""


class Scopus404Error(ScopusHtmlError):
    """Raised if a query yields a 404 error (Not Found for url)."""


class Scopus407Error(ScopusHtmlError):
    """Raised if a query yields a 407 error (Proxy Authentication Required)."""


class Scopus413Error(ScopusHtmlError):
    """Raised if a query yields a 413 error (Request Entity Too
    Large for url).
    """


class Scopus414Error(ScopusHtmlError):
    """Raised if a query yields a 414 error (Request-URI Too Large for url)."""


class Scopus429Error(ScopusHtmlError):
    """Raised if a query yields a 429 error (Quota exceeded)."""


class ScopusServerError(ScopusHtmlError):
    """Wrapper for Server related exceptions (code 5xx)."""
